from __future__ import annotations

import os
import tempfile
from pathlib import Path
from typing import Any
from typing import Dict
from typing import List

import h5py
import numpy
from ewoksutils.import_utils import qualname
from silx.io.dictdump import dicttoh5
from silx.io.dictdump import dicttonx
from silx.io.url import DataUrl
from silx.resources import ExternalResources

from darfix.core.dataset import ImageDataset

utilstest = ExternalResources(
    project="darfix",
    url_base="http://www.edna-site.org/pub/darfix/testimages",
    env_key="DATA_KEY",
    timeout=60,
)


def createHDF5Dataset(
    data: numpy.ndarray, metadata_dict: dict = None, output_file=None
):
    if metadata_dict is None:
        metadata_dict = {}
    if output_file is None:
        output_file = os.path.join(str(tempfile.mkdtemp()), "darfix_dataset.hdf5")
    dicttoh5(metadata_dict, output_file, h5path="1.1/instrument/positioners")
    with h5py.File(output_file, mode="a") as h5f:
        h5f["1.1/instrument/detector/data"] = data
    dataset = ImageDataset(
        _dir=os.path.dirname(output_file),
        detector_url=DataUrl(
            file_path=output_file,
            data_path="1.1/instrument/detector/data",
            scheme="silx",
        ),
        metadata_url=DataUrl(
            file_path=output_file,
            data_path="1.1/instrument/positioners",
            scheme="silx",
        ),
    )
    return dataset


def createHDF5Dataset1D(data: numpy.ndarray, output_file=None) -> ImageDataset:
    """
    Creates an ImageDataset with `data` inside and a single moving motor in the associated metadata (1D)

    :param data: 3D numpy array (N_motor_steps, N_pixels_x, N_pixels_y)
    :param output_file: Name of the output HDF5 file where the data will be saved. If None (default), the file will be saved in a temporary folder
    """
    assert data.ndim == 3

    N_frames = data.shape[0]
    metadata_dict = {"motor1": numpy.arange(N_frames)}

    return createHDF5Dataset(data, metadata_dict, output_file)


def createHDF5Dataset2D(data: numpy.ndarray, output_file=None) -> ImageDataset:
    """
    Creates an ImageDataset with `data` inside and a two moving motors in the associated metadata (2D)

    :param data: 4D dataset (N_motor1_steps, N_motor2_steps, N_pixels_x, N_pixels_y)
    :param output_file: Name of the output HDF5 file where the data will be saved. If None (default), the file will be saved in a temporary folder
    """
    assert data.ndim == 4

    N_motor1 = data.shape[0]
    N_motor2 = data.shape[1]
    N_frames = N_motor1 * N_motor2
    motor_1 = numpy.empty((N_motor1, N_motor2))
    motor_2 = numpy.empty((N_motor1, N_motor2))
    for i in range(N_motor1):
        motor_1[i] = numpy.arange(N_motor2)
    for j in range(N_motor2):
        motor_2[:, j] = numpy.arange(N_motor1)
    metadata_dict = {"motor1": motor_1.flatten(), "motor2": motor_2.flatten()}

    data = data.reshape((N_frames, *data.shape[2:]))

    return createHDF5Dataset(data, metadata_dict, output_file)


def createHDF5Dataset3D(data: numpy.ndarray, output_file=None) -> ImageDataset:
    """
    Creates an ImageDataset with `data` inside and a three moving motors in the associated metadata (3D)

    :param data: 5D dataset (N_motor1_steps, N_motor2_steps, N_motor3_steps, N_pixels_x, N_pixels_y)
    :param output_file: Name of the output HDF5 file where the data will be saved. If None (default), the file will be saved in a temporary folder
    """
    assert data.ndim == 5

    N_motor1, N_motor2, N_motor3 = data.shape[:3]
    N_frames = N_motor1 * N_motor2 * N_motor3
    motor_1 = numpy.empty((N_motor1, N_motor2, N_motor3))
    motor_2 = numpy.empty((N_motor1, N_motor2, N_motor3))
    motor_3 = numpy.empty((N_motor1, N_motor2, N_motor3))

    motor_1[:, :, :] = numpy.arange(N_motor1).reshape(N_motor1, 1, 1)
    motor_2[:, :, :] = numpy.arange(N_motor2).reshape(1, N_motor2, 1)
    motor_3[:, :, :] = numpy.arange(N_motor3).reshape(1, 1, N_motor3)
    metadata_dict = {
        "motor1": motor_1.flatten(),
        "motor2": motor_2.flatten(),
        "motor_3": motor_3.flatten(),
    }

    data = data.reshape((N_frames, *data.shape[-2:]))

    return createHDF5Dataset(data, metadata_dict, output_file)


def createRandomHDF5Dataset(
    dims,
    nb_data_frames=20,
    output_file=None,
    num_dims=3,
    metadata=False,
):
    """Simple creation of a dataset in output_file with the requested number of data
    files and dark files.

    :param tuple of int dims: dimensions of the files.
    :param int nb_data_frames: Number of data files to create.
    :param str or None output_file: output HDF5 file
    :param int num_dims: number of dimensions of the dataset

    :return :class:`Dataset`: generated instance of :class:`Dataset`
    """
    if not isinstance(dims, tuple) and len(dims) == 2:
        raise TypeError("dims should be a tuple of two elements")
    if not isinstance(nb_data_frames, int):
        raise TypeError(
            f"nb_data_frames ({nb_data_frames}) should be an int. Get {type(nb_data_frames)} instead"
        )
    if not isinstance(output_file, (type(None), str)):
        raise TypeError(
            f"output_file shuld be none or a string. Get {type(output_file)} instead"
        )

    if output_file is None:
        output_file = os.path.join(str(tempfile.mkdtemp()), "darfix_dataset.hdf5")

    metadata_dict = {}
    if metadata:
        metadata_dict["obx"] = [numpy.random.rand(1)[0]] * nb_data_frames
        metadata_dict["mainx"] = [numpy.random.rand(1)[0]] * nb_data_frames
        metadata_dict["ffz"] = [numpy.random.rand(1)[0]] * nb_data_frames
        metadata_dict["y"] = [numpy.random.rand(1)[0]] * nb_data_frames

        # comes from createRandomEDFDataset (legacy and remove in 4.0). Don't know why those
        # values are making sense...
        a = sorted(numpy.random.rand(2))
        b = [numpy.random.rand()] * numpy.array([1, 1.2, 1.4, 1.6, 1.8])
        c = sorted(numpy.random.rand(2))
        metadata_dict["m"] = [b[i % 5] for i in range(nb_data_frames)]
        if num_dims > 1:
            metadata_dict["z"] = [
                a[int((i > 4 and i < 10) or i > 14)] for i in range(nb_data_frames)
            ]
        metadata_dict["obpitch"] = [c[int(i > 9)] for i in range(nb_data_frames)]

    data = numpy.random.random((nb_data_frames, *dims))
    dicttoh5(metadata_dict, output_file, h5path="1.1/instrument/positioners")
    with h5py.File(output_file, mode="a") as h5f:
        h5f["1.1/instrument/detector/data"] = data
    assert os.path.exists(output_file)

    dataset = ImageDataset(
        _dir=os.path.dirname(output_file),
        detector_url=DataUrl(
            file_path=output_file,
            data_path="1.1/instrument/detector/data",
            scheme="silx",
        ),
        metadata_url=DataUrl(
            file_path=output_file,
            data_path="1.1/instrument/positioners",
            scheme="silx",
        ),
    )
    return dataset


def createDataset(
    data: numpy.ndarray, metadata_dict: dict = None, _dir: str | None = None
) -> ImageDataset:
    assert len(data) > 0
    if _dir is None:
        _dir = tempfile.mkdtemp()

    dir = Path(_dir)
    output_file = dir / "darfix_dataset.h5"
    return createHDF5Dataset(data, metadata_dict, output_file)


def create_scans(
    file_path: str,
    n_scan: int = 3,
    detector_path=r"{scan}/measurement/my_detector",
    metadata_path=r"{scan}/instrument/positioners",
):
    """
    create 'n_scan' scans with a detector like dataset and a 'positioners' groups containing motor like datasets

    warning: one of the dataset (delta) has an incoherent number of points (2 instead of 4). This is done on purpose
    to check behavior with this use case.
    """
    raw_detector_dataset = numpy.linspace(0, 5, 100 * 100 * 4).reshape(4, 100, 100)
    positioners_metadata = {
        "alpha": 1.0,
        "beta": numpy.arange(4, dtype=numpy.float32),
        "gamma": numpy.linspace(68, 70, 4, dtype=numpy.uint8),
        "delta": numpy.arange(2, dtype=numpy.int16),
    }

    for i in range(1, n_scan + 1):
        with h5py.File(file_path, mode="a") as h5f:
            h5f[detector_path.format(scan=f"{i}.1")] = raw_detector_dataset

        dicttonx(
            positioners_metadata,
            h5file=file_path,
            h5path=metadata_path.format(scan=f"{i}.1"),
            mode="a",
        )


def generate_ewoks_task_inputs(task_class, **kwargs) -> List[Dict[str, Any]]:
    task_identifier = qualname(task_class)

    return [
        {"task_identifier": task_identifier, "name": name, "value": value}
        for name, value in kwargs.items()
    ]


N_FRAMES_DIM0 = 10


def create_1d_dataset(dir, motor1, motor2):
    n_frames = N_FRAMES_DIM0
    dims = (n_frames, 100, 100)
    data = numpy.zeros(dims, dtype=numpy.float64)

    for i in range(n_frames):
        data[i] = i
    metadata_dict = {
        "mainx": numpy.ones(N_FRAMES_DIM0) * 0.5,
        motor1: numpy.ones(N_FRAMES_DIM0) * 0.2,
        motor2: numpy.arange(N_FRAMES_DIM0),
    }

    return createDataset(data, metadata_dict, dir)


def create_dataset_for_RSM(dir):
    """Create a dataset with suitable motor names to test RSM tasks"""
    return create_1d_dataset(dir, motor1="ffz", motor2="diffry")


def create_2_motors_metadata(
    slow_npoints: int, fast_npoints: int
) -> dict[str, numpy.ndarray]:
    """
    Create random metadata for 2 motors slow anf fast with
    """
    return {
        "slow": numpy.repeat(numpy.random.rand(slow_npoints), fast_npoints),
        "fast": numpy.tile(numpy.random.rand(fast_npoints), slow_npoints),
    }


def create_3_motors_metadata(
    slow_name: str,
    fast_name: str,
    very_fast_name: str,
    slow_npoints: int,
    fast_npoints: int,
    very_fast_npoints: int,
) -> dict[str, numpy.ndarray]:
    """
    Create random metadata for 2 motors slow anf fast with
    """
    return {
        slow_name: numpy.repeat(
            numpy.random.rand(slow_npoints), very_fast_npoints * fast_npoints
        ),
        fast_name: numpy.tile(
            numpy.repeat(numpy.random.rand(fast_npoints), slow_npoints),
            very_fast_npoints,
        ),
        very_fast_name: numpy.tile(
            numpy.random.rand(very_fast_npoints), slow_npoints * fast_npoints
        ),
        "mainx": 1,
        "obx": 1,
    }


def _3motors_dataset_args():
    """ "
    Creating random dataset with specific metadata.
    """
    dims = (20, 100, 100)
    data = numpy.zeros(dims)
    background = numpy.random.random(dims) * 100
    idxs = [0, 2, 4]
    data[idxs] += background[idxs]
    return data, create_3_motors_metadata("obpitch", "z", "m", 2, 2, 5)


def create_3motors_dataset(dir):
    """Create a dataset with 3 motors"""
    return createDataset(
        *_3motors_dataset_args(),
        _dir=str(dir) if dir else None,
    )



import os
from copy import deepcopy

import numpy
import pytest
from ewokscore.missing_data import MISSING_DATA

from darfix import dtypes
from darfix.tasks.dimensiondefinition import DimensionDefinition
from darfix.tests.utils import create_3_motors_metadata
from darfix.tests.utils import createDataset


@pytest.mark.parametrize("pre_compute_dims", (True, False))
def test_dimension_definition(tmp_path, pre_compute_dims):
    output_dir = os.path.join(tmp_path, "test_dimension_definition")
    os.makedirs(output_dir, exist_ok=True)

    metadata_dict = create_3_motors_metadata("obpitch", "z", "m", 2, 5, 2)
    data = numpy.zeros((20, 10, 100))
    darfix_dataset = createDataset(
        data=data, metadata_dict=metadata_dict, _dir=output_dir
    )

    if pre_compute_dims:
        darfix_dataset.find_dimensions()
        dataset_dims = deepcopy(darfix_dataset.dims)
    else:
        dataset_dims = MISSING_DATA

    dataset = dtypes.Dataset(dataset=darfix_dataset)

    task = DimensionDefinition(
        inputs={
            "dataset": dataset,
            "dims": dataset_dims,
        }
    )
    task.run()
    assert isinstance(task.outputs.dataset, dtypes.Dataset)
    assert len(task.outputs.dataset.dataset.dims) == 3
    assert task.outputs.dataset.dataset.dims[0].name == "m"
    assert task.outputs.dataset.dataset.dims[1].name == "z"
    assert task.outputs.dataset.dataset.dims[2].name == "obpitch"



from __future__ import annotations

import logging
import os.path

from ewokscore import Task
from ewokscore.missing_data import MISSING_DATA
from ewokscore.missing_data import MissingData
from ewokscore.model import BaseInputModel
from pydantic import ConfigDict
from pydantic import Field
from silx.io.url import DataUrl

from .. import dtypes
from ..core.data_selection import load_process_data

_logger = logging.getLogger(__file__)


class Inputs(BaseInputModel):
    model_config = ConfigDict(use_attribute_docstrings=True)
    raw_input_file: str = Field(
        examples=["/path/to/awesome/file.h5"], description="Path to the raw input file."
    )
    raw_detector_data_path: str = Field(
        examples=["/1.1/measurement/pco_ff"],
        description="Path to the raw detector data in the input file.",
    )
    raw_metadata_path: str | MissingData = Field(
        examples=["/1.1/instrument/positioners"],
        description="Path to the raw metadata in the input file.",
    )
    dark_input_file: str | MissingData = Field(
        default=MISSING_DATA,
        examples=["/path/to/dark/file.h5"],
        description="Path to the dark input file. Default is None.",
    )
    dark_detector_data_path: str | MissingData = Field(
        default=MISSING_DATA,
        examples=["/1.1/measurement/pco_ff"],
        description="Path to the dark detector data in the input file. Default is None.",
    )
    workflow_title: str | MissingData = MISSING_DATA
    """Title of the dataset for display purpose. Empty if not provided."""
    treated_data_dir: str | MissingData = MISSING_DATA
    """Processed output directory. If not provided, will try to find PROCESSED_DATA directory."""


class HDF5DataSelection(
    Task,
    input_model=Inputs,
    output_names=["dataset"],
):
    """Loads data and positioner metadata from a hdf5 file to create a Darfix dataset."""

    def run(self):
        raw_data_url = DataUrl(
            file_path=self.inputs.raw_input_file,
            data_path=self.inputs.raw_detector_data_path,
            scheme="silx",
        )

        raw_metadata_path = self.get_input_value("raw_metadata_path", None)
        if not raw_metadata_path:
            metadata_url = None
        else:
            metadata_url = DataUrl(
                file_path=self.inputs.raw_input_file,
                data_path=raw_metadata_path,
                scheme="silx",
            )

        dark_input_file = self.get_input_value("dark_input_file", None)
        dark_detector_data_path = self.get_input_value("dark_detector_data_path", None)
        if dark_input_file is None and dark_detector_data_path is not None:
            raise ValueError(
                "data path provided for background but no file path given."
            )

        if dark_input_file and dark_detector_data_path:
            bg_data_url = DataUrl(
                file_path=dark_input_file,
                data_path=dark_detector_data_path,
                scheme="silx",
            )
        else:
            bg_data_url = None

        treated_data_dir = self.get_input_value("treated_data_dir", None)
        # Move output folder to 'PROCESSED_DATA' instead of 'RAW_DATA' when possible
        # but ONLY if it was not set by the user
        if treated_data_dir is None and "RAW_DATA" in raw_data_url.file_path():
            treated_data_dir = os.path.dirname(raw_data_url.file_path()).replace(
                "RAW_DATA", "PROCESSED_DATA"
            )
            _logger.warning(f"Treated data will be saved in {treated_data_dir}")

        dataset, indices, bg_indices, bg_dataset = load_process_data(
            detector_url=raw_data_url.path(),
            root_dir=treated_data_dir,
            dark_detector_url=bg_data_url,
            title=self.get_input_value("workflow_title", ""),
            metadata_url=metadata_url,
        )

        self.outputs.dataset = dtypes.Dataset(
            dataset=dataset,
            indices=indices,
            bg_indices=bg_indices,
            bg_dataset=bg_dataset,
        )



from __future__ import annotations

import logging
from typing import Any
from typing import Iterable

import numpy
from ewokscore import Task
from ewokscore.missing_data import MISSING_DATA
from ewokscore.missing_data import MissingData
from ewokscore.model import BaseInputModel
from pydantic import ConfigDict
from pydantic import Field

from darfix import dtypes
from darfix.core.dataset import ImageDataset
from darfix.core.dimension import AcquisitionDims
from darfix.core.dimension import Dimension
from darfix.core.dimension import find_dimensions_from_metadata
from darfix.core.fscan_parser import fscan_get_dimensions
from darfix.core.zigzag_mode import reorder_frames_of_zigzag_scan

_logger = logging.getLogger(__file__)

REL_TOLERANCE_FOR_MIN_MAX_VALIDATION = 1e-1  # 10%


class Inputs(BaseInputModel):
    model_config = ConfigDict(use_attribute_docstrings=True)
    dataset: dtypes.Dataset
    """ Input dataset containing a stack of images """
    dims: dict[int, Any] | MissingData = Field(
        default=MISSING_DATA,
        examples=[
            {
                0: {"name": "diffrx", "size": 5, "range": [0.0, 5.0, 1.0]},
                1: {"name": "diffry", "size": 10, "range": [0.0, 10.0, 1.0]},
            }
        ],
        description="Dimensions to use for the dataset. If not provided, the task will try to find dimensions from metadata.",
    )
    tolerance: float | MissingData = MISSING_DATA
    """Tolerance to use for finding dimensions from metadata. Default is 1e-9."""
    is_zigzag: bool | MissingData = MISSING_DATA
    """Set to True if the scan was a zigzag scan (slow motor moving back and forth). Defaults to False."""


class DimensionDefinition(
    Task,
    input_model=Inputs,
    output_names=["dataset"],
):
    """
    Fit dimension of given dataset.
    If dims are provided then will use them. else will call 'find_dimensions' with the provided tolerance or the default one.
    """

    DEFAULT_TOLERANCE = 1e-9

    def run(self):
        if not isinstance(self.inputs.dataset, dtypes.Dataset):
            raise TypeError(
                f"'dataset' input should be an instance of {dtypes.Dataset}. Got {type(self.inputs.dataset)}"
            )

        dataset = self.inputs.dataset.dataset
        if not isinstance(dataset, dtypes.ImageDataset):
            raise TypeError(
                f"self.inputs.dataset is expected to be an instance of {dtypes.ImageDataset}. Get {type(dataset)}"
            )

        fscan_parameters = fscan_get_dimensions(dataset)

        dims = self._handle_dims(
            fscan_parameters[1] if fscan_parameters else None,
            dataset.metadata_dict,
            user_input=self._get_dimensions_from_default_inputs(),
        )
        is_zigzag = self._handle_is_zigzag(
            fscan_parameters[0] if fscan_parameters else None,
            user_input=self.get_input_value("is_zigzag", None),
        )

        assert_dimensions_ok(dataset, dims.values())

        if is_zigzag:
            reorder_frames_of_zigzag_scan(dims, dataset)

        # Reshape the dataset with the new dimensions

        dataset.dims = dims
        dataset.reshape_data()
        self.outputs.dataset = dtypes.Dataset(
            dataset=dataset,
            indices=self.inputs.dataset.indices,
            bg_indices=self.inputs.dataset.bg_indices,
            bg_dataset=self.inputs.dataset.bg_dataset,
        )

    def _get_dimensions_from_default_inputs(self) -> AcquisitionDims | None:

        raw_dims = self.get_input_value("dims", None)

        if raw_dims is None:
            return None

        try:
            return AcquisitionDims.from_dict(raw_dims)
        except ValueError as e:
            # TODO: Should we really silence the error here?
            _logger.error(f"Encountered {e} when parsing default raw_dims: {raw_dims}")
            return None

    def _handle_is_zigzag(self, fscan_input: bool | None, user_input: bool | None):
        if user_input:
            _logger.debug("is_zigzag set by user")
            return user_input

        if fscan_input:
            _logger.debug("is_zigzag set by fscan")
            return fscan_input

        _logger.debug("Using default value (False) for is_zigzag")
        return False

    def _handle_dims(
        self,
        fscan_dims: dict[int, Dimension] | None,
        metadata_dict: dict,
        user_input: AcquisitionDims | None,
    ) -> AcquisitionDims:
        if user_input is not None:
            _logger.debug("dims set by user")
            return user_input

        if fscan_dims is not None:
            _logger.debug("dims set by fscan")
            return AcquisitionDims.from_dict(fscan_dims)

        _logger.debug("dims computed from metadata")
        return self._compute_dimensions_from_metadata(metadata_dict)

    def _compute_dimensions_from_metadata(
        self, metadata_dict: dict[str, numpy.ndarray]
    ) -> AcquisitionDims:

        tolerance = self.get_input_value("tolerance", self.DEFAULT_TOLERANCE)

        return find_dimensions_from_metadata(
            metadata_dict,
            tolerance,
        )


def assert_dimensions_ok(dataset: ImageDataset, dims: Iterable[Dimension]) -> None:

    error_msg = get_dimensions_error(dataset, dims)
    if error_msg:
        raise RuntimeError(error_msg)


def get_dimensions_error(
    dataset: ImageDataset, dims: Iterable[Dimension]
) -> str | None:

    shape = []

    # Check name, min and max
    for dimension in dims:

        shape.append(dimension.size)

        if dimension.name not in dataset.metadata_dict:
            return f"Dimension with name '{dimension.name}' is not in the metadata."
        metadata_min = numpy.min(dataset.metadata_dict[dimension.name])
        metadata_max = numpy.max(dataset.metadata_dict[dimension.name])
        rel_tol = (metadata_max - metadata_min) * REL_TOLERANCE_FOR_MIN_MAX_VALIDATION
        metadata_min -= rel_tol
        metadata_max += rel_tol

        if dimension.start > metadata_max or dimension.start < metadata_min:
            return f"Dimension with name '{dimension.name}' start value = {dimension.start} but this is outside the dimension range [{metadata_min} → {metadata_max}]."

        if dimension.stop > metadata_max or dimension.stop < metadata_min:
            return f"Dimension with name '{dimension.name}' stop value = {dimension.stop} but this is outside the dimension range [{metadata_min} → {metadata_max}]."

    if len(shape) == 0:
        return "None dimension are defined."

    dims_size = numpy.prod(shape)

    # Check size
    if dims_size != dataset.nframes:
        product_string = " x ".join([str(size) for size in shape])

        return f"Dimensions do not match number of frames → {product_string} ≠ {dataset.nframes}."

    return None



from silx.gui import qt
from silx.gui.colors import Colormap
from silx.gui.plot.StackView import StackViewMainWindow

from .. import config
from ..dtypes import ImageDataset
from .chooseDimensions import ChooseDimensionWidget


class ProjectionWidget(qt.QWidget):
    """
    Widget to apply a projection to the chosen dimension.
    """

    sigOkClicked = qt.Signal()
    sigProjectButtonClicked = qt.Signal()
    sigDimensionsChanged = qt.Signal()

    def __init__(self, parent=None):
        super().__init__(parent=parent)

        self._sv = StackViewMainWindow()
        self._sv.setColormap(
            Colormap(name=config.DEFAULT_COLORMAP_NAME, normalization="linear")
        )
        self._sv.setKeepDataAspectRatio(True)
        self._chooseDimensionWidget = ChooseDimensionWidget(
            self, vertical=False, values=False, _filter=False
        )
        self._projectButton = qt.QPushButton("Project data")
        self._projectButton.setEnabled(False)
        types = qt.QDialogButtonBox.Ok
        self._buttons = qt.QDialogButtonBox(parent=self)
        self._buttons.setStandardButtons(types)
        self._buttons.setEnabled(False)
        layout = qt.QGridLayout()
        layout.addWidget(self._chooseDimensionWidget, 0, 0, 1, 2)
        layout.addWidget(self._projectButton, 1, 1)
        layout.addWidget(self._sv, 2, 0, 1, 2)
        layout.addWidget(self._buttons, 3, 1)
        self.setLayout(layout)

        self._projectButton.clicked.connect(self.sigProjectButtonClicked.emit)
        self._buttons.accepted.connect(self.sigOkClicked.emit)
        self._chooseDimensionWidget.valueChanged.connect(self.sigDimensionsChanged.emit)

    def setDataset(self, dataset: ImageDataset):
        if not dataset.dims:
            msg = qt.QMessageBox()
            msg.setIcon(qt.QMessageBox.Warning)
            msg.setText(
                "This widget has to be used after selecting the dimensions of the dataset. Run the `Dimension definition` widget first."
            )
            msg.exec()
            return

        self._projectButton.setEnabled(True)
        if dataset.dims.ndim > 1:
            self._buttons.setEnabled(True)

        self._chooseDimensionWidget.setDimensions(dataset.dims)
        self._chooseDimensionWidget._updateState(True)
        for i in range(1, dataset.dims.ndim - 1):
            self._chooseDimensionWidget.dimensionWidgets[i][0].addItem("None")
        self._sv.setKeepDataAspectRatio(True)
        self._sv.setGraphTitle(dataset.title)

    def getDimension(self):
        dimension = [self._chooseDimensionWidget.dimension[0]]
        for i in range(1, len(self._chooseDimensionWidget.dimension)):
            dimension += [self._chooseDimensionWidget.dimension[i]]

        return dimension

    def updatePlot(self, dataset: ImageDataset):
        self._projectButton.setEnabled(True)
        self._sv.setStack(dataset.as_array3d())



from __future__ import annotations

from silx.gui import qt


class VSpacer(qt.QWidget):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setLayout(qt.QVBoxLayout())
        self.setSizePolicy(
            qt.QSizePolicy.Policy.Minimum, qt.QSizePolicy.Policy.Expanding
        )



import numpy
import pytest

from darfix.core.transformation import Transformation
from darfix.core.utils import NoDimensionsError

from .utils import create_1d_dataset
from .utils import create_dataset_for_RSM


def test_rsm_kind(tmpdir):
    dataset = create_dataset_for_RSM(dir=tmpdir)

    with pytest.raises(NoDimensionsError):
        dataset.compute_transformation(0.1, kind="rsm")
    dataset.find_dimensions()

    assert dataset.transformation is None
    dataset.compute_transformation(0.1, kind="rsm")

    transformation = dataset.transformation

    assert isinstance(transformation, Transformation)
    assert transformation.shape == dataset.frame_shape
    assert numpy.all(numpy.isfinite(transformation.x))
    assert numpy.all(numpy.isfinite(transformation.y))


def test_magnification_kind(tmpdir):
    dataset = create_1d_dataset(
        dir=tmpdir,
        motor1="obx",
        motor2="obpitch",
    )

    with pytest.raises(NoDimensionsError):
        dataset.compute_transformation(0.1, kind="magnification")
    dataset.find_dimensions()

    assert dataset.transformation is None
    dataset.compute_transformation(0.1, kind="magnification")

    transformation = dataset.transformation

    assert isinstance(transformation, Transformation)
    assert transformation.shape == dataset.frame_shape
    assert numpy.all(numpy.isfinite(transformation.x))
    assert numpy.all(numpy.isfinite(transformation.y))


def test_compute_magnification(dataset):
    """Tests fitting data"""

    dataset.find_dimensions()
    dataset.reshape_data()
    dataset.compute_transformation(d=0.1)
    assert dataset.transformation.shape == dataset.frame_shape



from __future__ import annotations

from typing import Any
from typing import List
from typing import Literal
from typing import Optional
from typing import Tuple
from typing import Union

import numpy

from darfix.core.dataset import ImageDataset

AxisAndValueIndices = Tuple[List[int], List[int]]

AxisType = Union[Literal["dims"], Literal["center"], None]


class Dataset:
    def __init__(
        self,
        dataset: ImageDataset,
        indices: Optional[numpy.ndarray] = None,
        bg_indices: Optional[numpy.ndarray] = None,
        bg_dataset: Optional[ImageDataset] = None,
    ):
        """Darfix dataset with indices and background

        :param dataset: Darfix dataset object that holds the image stack
        :param indices: Image stack indices to be taking into account. Usually set by the 'partition data' task. Defaults to None.
        :param bg_indices: Dark image stack indices to be taking into account. Usually set by the 'partition data' task. Defaults to None.
        :param bg_dataset: Darfix dataset object that holds the dark image stack. Defaults to None.
        """
        self.dataset = dataset
        self.indices = indices
        self.bg_indices = bg_indices
        self.bg_dataset = bg_dataset


class DatasetTypeError(TypeError):
    def __init__(self, wrong_dataset: Any):
        """Error raised when a dataset has not the expected Dataset type"""
        super().__init__(
            f"Dataset is expected to be an instance of {Dataset}. Got {type(wrong_dataset)}."
        )



import numpy
import pytest

from darfix.decomposition.base import Base


@pytest.fixture
def images():
    return numpy.random.random((100, 1000))


def test_data(images):
    base = Base(images)
    numpy.testing.assert_equal(images, base.data)


def test_indices(images):
    base = Base(images)
    numpy.testing.assert_equal(base.indices, numpy.arange(len(images)))

    base = Base(images, indices=numpy.arange(20))
    numpy.testing.assert_equal(base.indices, numpy.arange(20))


def test_num_components(images):
    base = Base(images)
    assert base.num_components == 100

    base = Base(images, num_components=10)
    assert base.num_components == 10


def test_W(images):
    base = Base(images, num_components=10)
    base.fit_transform()

    assert base.W.shape == (100, 10)


def test_H(images):
    base = Base(images, num_components=10)
    base.fit_transform()

    assert base.H.shape == (10, 1000)


def test_fit_transform(images):
    base = Base(images)
    base.fit_transform(compute_w=False)
    assert not hasattr(base, "W")
    assert hasattr(base, "H")

    base = Base(images)
    base.fit_transform(compute_h=False)
    assert not hasattr(base, "H")
    assert hasattr(base, "W")


def test_frobenius_norm(images):
    base = Base(images)
    assert base.frobenius_norm() is None

    base.fit_transform()
    assert base.frobenius_norm() is not None



from __future__ import annotations

import logging
from typing import Any
from typing import Optional

from ewokscore.missing_data import is_missing_data
from ewoksorange.bindings.owwidgets import OWEwoksWidgetOneThread
from ewoksorange.gui.parameterform import block_signals
from silx.gui import qt

from darfix import dtypes
from darfix.core.dataset import ImageDataset
from darfix.core.dimension import Dimension
from darfix.core.dimension import find_dimensions_from_metadata
from darfix.core.fscan_parser import fscan_get_dimensions
from darfix.gui.dimensionsWidget import DimensionItem
from darfix.gui.dimensionsWidget import DimensionWidget
from darfix.tasks.dimensiondefinition import DimensionDefinition
from darfix.tasks.dimensiondefinition import get_dimensions_error

_logger = logging.getLogger(__file__)

NO_DATASET_ERROR_MSG = "No dataset. Please run data selection widget."


class DimensionWidgetOW(OWEwoksWidgetOneThread, ewokstaskclass=DimensionDefinition):
    """
    Widget used to define the calibration of the experimentation (select motor
    positions...)
    """

    name = "Dimension definition"
    id = "orange.widgets.darfix.dimensiondefinition"
    description = "Define the motor axes used during the acquisition"
    icon = "icons/param_dims.svg"

    _ewoks_inputs_to_hide_from_orange = ("dims", "tolerance", "is_zigzag")

    priority = 4
    keywords = ["dataset", "calibration", "motor", "angle", "geometry"]
    want_main_area = False

    def __init__(self):
        super().__init__()
        self._widget = DimensionWidget(parent=self)
        self.controlArea.layout().addWidget(self._widget)

        # buttons
        types = qt.QDialogButtonBox.Ok
        self.buttons = qt.QDialogButtonBox(parent=self)
        self.buttons.setStandardButtons(types)
        self.controlArea.layout().addWidget(self.buttons)

        self.buttons.accepted.connect(self._execute_fit)
        self.buttons.button(qt.QDialogButtonBox.Ok).setEnabled(False)

        # connect Signal/SLOT
        self._widget.sigFindDimensions.connect(self._findDimensions)
        self._widget.sigUpdateDimensions.connect(self.onUpdateDimensions)
        # set up
        self._initDims()

        tolerance = self.get_task_input_value("tolerance")
        if not is_missing_data(tolerance):
            with block_signals(self._widget):
                self._widget.setTolerance(tolerance)

    def onUpdateDimensions(self):
        dataset = self.get_task_input_value("dataset")
        if is_missing_data(dataset):
            return

        err_msg = get_dimensions_error(
            dataset.dataset,
            (dim.toDimension() for dim in self._widget.dims.values()),
        )
        self.buttons.button(qt.QDialogButtonBox.Ok).setEnabled((err_msg is None))
        if err_msg is None:
            self._widget.setTipsLabelText(
                "\u2705 Dimensions are good. Confirm with 'OK' button to go to next step."
            )
        else:
            self._widget.setTipsLabelText("\u274c " + err_msg)

    def setDataset(self, dataset: Optional[dtypes.Dataset], pop_up: bool = False):
        """
        Input signal to set the dataset.
        """
        if dataset is None:
            return

        # TODO : How we handle zigzag mode need to be review somehow because as we change the dataset and metadata
        # but we do not make a copy, i cannot rerun twice the task
        # So that's a patch for now
        self.Information.clear()
        self.controlArea.setEnabled(True)

        fscan_parameters = fscan_get_dimensions(dataset.dataset)
        if fscan_parameters is not None:
            is_zigzag, dims = fscan_parameters
            self._widget.setDims(dims)
            self._widget.setZigzagMode(is_zigzag)

            fscan_err_msg = get_dimensions_error(dataset.dataset, dims.values())
            if fscan_err_msg is None:
                self._widget.setTipsLabelText(
                    "\u2705 Dimensions are defined in metadata. All is good ! Confirm with 'OK' button to go to next step."
                )
                # Disable inputs only when the dim rows are added
                self._widget.setEnableInputs(False)
            else:
                self._widget.setEnableInputs(True)
                _logger.error(
                    f"Find fscan_parameters in hdf5 file but validation failed : {fscan_err_msg}"
                )
        else:
            err_msg = get_dimensions_error(
                dataset.dataset,
                (dim.toDimension() for dim in self._widget.dims.values()),
            )
            self._widget.setEnableInputs(True)

            if err_msg is None:
                self.buttons.button(qt.QDialogButtonBox.Ok).setEnabled(True)
                self._widget.setTipsLabelText(
                    "\u2705 Dimensions are good. Confirm with 'OK' button to go to next step."
                )
            else:
                self._widget.setZigzagMode(False)
                self._widget.setDims({})
                self._widget.setTipsLabelText(
                    "\U0001f4a1 Adjust tolerance and click on 'Find dimensions'."
                )

        # note: set_dynamic_input instead of set_default_input to make sure the dataset
        # will not be saved as orange Settings in the .ows file
        self.set_dynamic_input("dataset", dataset)
        if pop_up:
            self.open()

    def closeEvent(self, event):
        self._save()

    def _initDims(self):
        try:
            # dims values are Dimension provided as dict (to save settings in a readable manner).
            # So when loading them we must convert them back to Dimension
            raw_dims = self.get_task_input_value("dims")
            if not is_missing_data(raw_dims):
                dims = convert_dim_from_dict_to_Dimension(raw_dims)
                with block_signals(self._widget):
                    self._widget.setDims(dims)
            self._widget.setZigzagMode(self.get_task_input_value("is_zigzag", False))
            self._widget.setEnableInputs(False)
        except ValueError as e:
            qt.QMessageBox.warning(self, "Fail to setup dimension definition", str(e))

    def _save(self):
        pickable_dims = make_dims_picklable(self._widget.dims)
        self.set_default_input("dims", pickable_dims)
        self.set_default_input("is_zigzag", self._widget.isZigzagMode)
        self.set_default_input("tolerance", self._widget.getTolerance())

    def _execute_fit(self):
        if not is_missing_data(self.get_task_input_value("dataset")):
            self._save()
            self.execute_ewoks_task_without_propagation()
        else:
            self._provideFitFeedback(
                False,
                error=NO_DATASET_ERROR_MSG,
                dataset=None,
            )

    def task_output_changed(self) -> None:
        dataset = self.get_task_output_value("dataset")
        if is_missing_data(dataset):
            darfix_dataset = None
        else:
            if not isinstance(dataset, dtypes.Dataset):
                raise dtypes.DatasetTypeError(dataset)
            darfix_dataset = dataset.dataset
        self._provideFitFeedback(
            success=self.task_succeeded,
            error=self.task_exception,
            dataset=darfix_dataset,
        )

        return super().task_output_changed()

    def _provideFitFeedback(self, success: bool, error: str, dataset):
        msg = qt.QMessageBox()
        if success:

            # TODO : How we handle zigzag mode need to be review somehow because as we change the dataset and metadata
            # but we do not make a copy, i cannot rerun twice the task
            # So that's a patch for now
            self.controlArea.setDisabled(True)
            self.information(
                "Dimension can be defined only once. If you want to redo, re-run previous data selection widget."
            )

            self.propagate_downstream()
            self.accept()
        else:
            msg.setIcon(qt.QMessageBox.Warning)
            msg.setText(f"Error: {error}")
            msg.setWindowTitle("Dimension fit failed!")
            msg.setStandardButtons(qt.QMessageBox.Ok)
            msg.exec()

    def handleNewSignals(self) -> None:
        """
        Today the DimensionWidgetOW is not processing automatically a dataset when it receives it.
        It wait the user to press optionally 'find dimension', then 'fit' and 'ok' to validate the task
        """
        dataset = self.get_task_input_value("dataset")
        if not is_missing_data(dataset):
            self.setDataset(dataset=dataset, pop_up=True)
        # return super().handleNewSignals() do not call to make sure the processing is not triggered

    def _findDimensions(
        self,
    ):
        dataset = self.get_task_input_value("dataset")
        if is_missing_data(dataset):
            _logger.error(NO_DATASET_ERROR_MSG)
            return
        assert isinstance(dataset.dataset, ImageDataset)

        metadata = dataset.dataset.metadata_dict
        dims = find_dimensions_from_metadata(metadata, self._widget.getTolerance())
        self._widget.setDims(dims)


def make_dims_picklable(dims: dict[DimensionItem]):
    if not isinstance(dims, dict):
        raise TypeError("dims should be an instance of dict")
    # convert AcquisitionDims to dict if necessary
    return {value.axis: make_dim_picklable(value) for key, value in dims.items()}


def make_dim_picklable(dim: DimensionItem) -> dict:
    return dim.toDimension().to_dict()


def convert_dim_from_dict_to_Dimension(
    dims: dict[int, dict[int, Any]],
) -> dict[int, Dimension]:
    return {key: Dimension.from_dict(value) for key, value in dims.items()}



from __future__ import annotations

import numpy
from silx.gui import qt
from silx.gui.plot import Plot1D

from darfix import dtypes

from .operationThread import OperationThread


class PCAPlot(Plot1D):
    def __init__(self, parent=None, backend=None):
        super().__init__(parent, backend)
        self.setDataMargins(0.05, 0.05, 0.05, 0.05)
        self.setGraphXLabel("Components")
        self.setGraphYLabel("Singular values")

    def setData(self, vals: numpy.ndarray, title=None):
        if not isinstance(vals, numpy.ndarray):
            raise TypeError(
                f"vals should be an instance of {numpy.ndarray}. got {type(vals)} instead"
            )
        if title is None:
            title = ""
        self.setGraphTitle("Components representation of the dataset " + title)
        self.addCurve(numpy.arange(len(vals)), vals, symbol=".", linestyle=" ")


class PCAWidget(qt.QMainWindow):
    """
    Widget to apply PCA to a set of images and plot the eigenvalues found.
    """

    sigComputed = qt.Signal()

    def __init__(self, parent=None):
        qt.QWidget.__init__(self, parent)

        self._plot = PCAPlot()
        self.setCentralWidget(self._plot)

    def _computePCA(self):
        try:
            self._thread = OperationThread(self, self._dataset.pca)
            self._thread.setArgs(return_vals=True, indices=self.indices)
            self._thread.finished.connect(self._updateData)
            self._thread.start()
        except Exception as e:
            raise e

    def setDataset(self, dataset: dtypes.Dataset):
        self._dataset = dataset.dataset
        self.indices = dataset.indices
        self.bg_indices = dataset.bg_indices
        self.bg_dataset = dataset.bg_dataset
        self._computePCA()

    def _updateData(self):
        """
        Plots the eigenvalues.
        """
        self._thread.finished.disconnect(self._updateData)
        vals = self._thread.data
        self._plot.setData(vals, self._dataset.title)
        self.sigComputed.emit()



from __future__ import annotations

import numpy
from silx.image.marchingsquares import find_contours


def compute_contours(image: numpy.ndarray) -> list[numpy.ndarray]:
    polygons = []
    for i in numpy.linspace(numpy.min(image), numpy.max(image), 10):
        polygons.extend(find_contours(image, i))

    return polygons



import numpy

from darfix.core.dataset import ImageDataset

from . import utils


def test_load_and_save():
    imgs = numpy.arange(1000).reshape((5, 2, 10, 10))
    dset = utils.createHDF5Dataset2D(imgs)
    dset.find_dimensions()
    dset.save_in_output_dir()
    dset2 = ImageDataset.load_from_output_dir(dset.dir)

    assert dset2.dims.ndim == 2
    assert dset2.dims[0].name == "motor1"
    assert dset2.dims[1].name == "motor2"
    assert "motor1" in dset2.metadata_dict
    assert "motor2" in dset2.metadata_dict
    numpy.testing.assert_allclose(dset2.data, imgs)



import numpy

from darfix.math import bivariate_gaussian
from darfix.math import gaussian
from darfix.math import trivariate_gaussian


def test_gaussian():
    X = numpy.random.random((100))
    X0 = X.mean()

    result = gaussian(X, amplitude=1, x0=X0, std_dev=0.5, background=0)

    assert result.shape == (100,)


def test_bivariate_gaussian():
    X = numpy.random.random((2, 100))
    X0 = X.mean(axis=1)

    result = bivariate_gaussian(X, X0[0], X0[1], 0.5, 0.5, 1, 0.1)

    assert result.shape == (100,)


def test_trivariate_gaussian():
    X = numpy.random.random((3, 100))
    X0 = X.mean(axis=1)

    result = trivariate_gaussian(
        X,
        X0[0],
        X0[1],
        X0[2],
        sigma_x0=1,
        sigma_x1=1,
        sigma_x2=1,
        c10=0.5,
        c20=0.2,
        c12=0.1,
        amplitude=1,
    )

    assert result.shape == (100,)



import numpy
from scipy import stats

import darfix
from darfix.core import mapping


def test_moments():
    """Tests the correct moments calculation"""
    weights = list()
    moments = list()
    rs = numpy.random.RandomState(100)

    n0 = 200
    n1 = 3
    n2 = 4
    x = numpy.linspace(-2, 6, n0)
    aparams = rs.uniform(-5, 5, n1 * n2)
    scales = rs.uniform(0.5, 1, n1 * n2)
    locs = rs.uniform(1, 3, n1 * n2)

    for loc, scale, a in zip(locs, scales, aparams):
        rv = stats.skewnorm(a=a, loc=loc, scale=scale)
        w = rv.pdf(x)

        # Theoretical
        mean0, var0, skew0, kurt0 = rv.stats(moments="mvsk")

        # Calculate from data
        mean = sum(w * x) / w.sum()
        resid = x - mean
        var = sum(w * resid**2) / w.sum()
        sigma = numpy.sqrt(var)
        resid /= sigma
        fwhm = sigma * darfix.config.FWHM_VAL
        skew = sum(w * resid**3) / w.sum()
        kurt = (sum(w * resid**4) / w.sum()) - 3

        # Compare
        numpy.testing.assert_allclose(mean, mean0, rtol=1e-4)
        numpy.testing.assert_allclose(var, var0, rtol=1e-3)
        numpy.testing.assert_allclose(skew, skew0, rtol=1e-1)
        numpy.testing.assert_allclose(kurt, kurt0, rtol=2e-1)

        moments.append([mean, fwhm, skew, kurt])
        weights.append(w)

    weights = numpy.asarray(weights)
    weights = weights.reshape((n1, n2, n0))
    weights = numpy.transpose(weights, (2, 0, 1))
    moments = numpy.asarray(moments).reshape((n1, n2, 4))
    moments = numpy.transpose(moments, (2, 0, 1))
    mean1, fwhm1, skew1, kurt1 = moments

    mean2, fwhm2, skew2, kurt2 = mapping.compute_moments(x, weights, smooth=False)

    numpy.testing.assert_allclose(mean1, mean2)
    numpy.testing.assert_allclose(fwhm1, fwhm2)
    numpy.testing.assert_allclose(skew1, skew2)
    numpy.testing.assert_allclose(kurt1, kurt2)

    if False:
        import matplotlib.pyplot as plt

        for i in range(n1):
            for j in range(n2):
                y = weights[:, i, j]
                plt.plot(x, y)
                plt.vlines(mean1[i, j], 0, y.max())
                plt.vlines(mean2[i, j], 0, y.max())
        plt.show()


def test_rsm():
    """Tests RSM"""
    data = numpy.random.random(size=(3, 10, 10))
    H, W = data.shape[1:]
    d = 0.1
    ffz = 10
    mainx = 5

    pix_arr = mapping.compute_rsm(H, W, d, ffz, mainx)

    assert pix_arr[0].shape == (H, W)
    assert pix_arr[1].shape == (H, W)


def test_magnification():
    """Tests magnification"""
    data = numpy.random.random(size=(3, 10, 10))
    H, W = data.shape[1:]
    d = 0.1
    obx = 10
    obpitch = 25.1
    mainx = 5

    pix_arr = mapping.compute_magnification(H, W, d, obx, obpitch, mainx)

    assert pix_arr[0].shape == (H, W)
    assert pix_arr[1].shape == (H, W)


def test_magnification_uncentered():
    """Tests magnification uncentered"""
    data = numpy.random.random(size=(3, 10, 10))
    H, W = data.shape[1:]
    d = 0.1
    obx = 10
    obpitch = 25.1
    mainx = 50  # Has to be big enough

    pix_arr = mapping.compute_magnification(H, W, d, obx, obpitch, mainx, center=False)

    assert pix_arr[0][0][0] == 0
    assert pix_arr[1][H - 1][0] == 0


def test_peak_position():
    """Tests peak position map"""
    data = numpy.random.random(size=(3, 10, 10))
    image = mapping.compute_peak_position(data)
    assert image[0, 1] == numpy.argmax(data[:, 0, 1])


def test_peak_position_values():
    """Tests peak position map with values"""
    data = numpy.random.random(size=(3, 10, 10))
    values = numpy.repeat([numpy.linspace(0.1, 1, 10)], 3).flatten()
    image = mapping.compute_peak_position(data, values)

    assert image[0, 0] == values[numpy.argmax(data[:, 0, 0])]



from __future__ import annotations

import numpy
from ewoksorange.gui.parameterform import block_signals
from silx.gui import qt
from silx.gui.colors import Colormap
from silx.gui.plot import Plot2D
from silx.io.dictdump import dicttonx

from ... import dtypes
from ...core.dataset import Operation
from ...core.rocking_curves import MAPS_1D
from ...core.rocking_curves import MAPS_2D
from ...core.rocking_curves import generate_rocking_curves_nxdict
from ...core.utils import NoDimensionsError
from ...core.utils import TooManyDimensionsForRockingCurvesError
from ..utils.message import missing_dataset_msg
from ..utils.utils import select_output_hdf5_file_with_dialog
from .fitComboBox import FitComboBox
from .rockingCurvesPlot import RockingCurvesPlot

# "Residuals" are not given by the fit but computed by the widget.
# It needs to be handled separately of other `MAPS` values
MAPS_CB_OPTIONS_1D = [*MAPS_1D, "Residuals"]
MAPS_CB_OPTIONS_2D = [*MAPS_2D, "Residuals"]


def _get_option_label(item: str, dataset: dtypes.ImageDataset):
    if "first motor" in item:
        return item.replace("first motor", dataset.dims.get(0).name)

    if "second motor" in item:
        return item.replace("second motor", dataset.dims.get(1).name)

    return item


class RockingCurvesWidget(qt.QWidget):
    """
    Widget to apply fit to a set of images and plot the amplitude, fwhm, peak position, background and residuals maps.
    """

    sigFitClicked = qt.Signal()

    def __init__(self, parent=None):
        qt.QWidget.__init__(self, parent)

        self.dataset = None
        self.indices = None
        self._update_dataset = None
        self._residuals_cache = None
        self.maps = None
        """ Holds the result of the fit as as a stack of maps"""

        layout = qt.QVBoxLayout()

        self._rockingCurvesPlot = RockingCurvesPlot(parent=self)
        layout.addWidget(self._rockingCurvesPlot)

        fitLayout = qt.QHBoxLayout()
        self._intensityThresholdLE = qt.QLineEdit("15")
        self._intensityThresholdLE.setValidator(qt.QDoubleValidator())
        fitLayout.addWidget(qt.QLabel("Intensity threshold:"))
        fitLayout.addWidget(self._intensityThresholdLE)

        self._fitMethodCB = FitComboBox()
        fitLayout.addWidget(qt.QLabel("Method"))
        fitLayout.addWidget(self._fitMethodCB)

        self._launchFitBtn = qt.QPushButton("Fit data")
        fitLayout.addWidget(self._launchFitBtn)
        self._launchFitBtn.clicked.connect(self._launchFit)
        layout.addLayout(fitLayout)

        self._abortFitBtn = qt.QPushButton("Abort")
        self._abortFitBtn.hide()
        self._abortFitBtn.clicked.connect(self.__abort)
        layout.addWidget(self._abortFitBtn)

        self._mapsCB = qt.QComboBox(self)
        self._mapsCB.hide()
        layout.addWidget(self._mapsCB)
        self._plotMaps = Plot2D(self)
        self._plotMaps.setDefaultColormap(
            Colormap(name="cividis", normalization="linear")
        )
        self._plotMaps.hide()
        layout.addWidget(self._plotMaps)
        self._exportButton = qt.QPushButton("Export maps")
        self._exportButton.hide()
        self._exportButton.clicked.connect(self.exportMaps)
        layout.addWidget(self._exportButton)

        self.setLayout(layout)

        # connect signal / slot
        self._mapsCB.currentTextChanged.connect(self._displayMap)

    def setDataset(self, dataset: dtypes.Dataset):
        if not dataset.dataset.dims.ndim:
            raise NoDimensionsError("RockingCurvesWidget")
        self.dataset = dataset.dataset
        self.indices = dataset.indices
        self._update_dataset = dataset.dataset
        self._residuals_cache = None

        if self.dataset.dims.ndim == 1:
            options = MAPS_CB_OPTIONS_1D
        elif self.dataset.dims.ndim == 2:
            options = MAPS_CB_OPTIONS_2D
        else:
            raise TooManyDimensionsForRockingCurvesError()

        with block_signals(self._mapsCB):
            self._mapsCB.clear()
            for option in options:
                self._mapsCB.addItem(_get_option_label(option, dataset=dataset.dataset))

        self._rockingCurvesPlot.setDataset(dataset)
        self._rockingCurvesPlot.updateStack()

    def _launchFit(self):
        """
        Method called when button for computing fit is clicked
        """
        if self.dataset is None:
            missing_dataset_msg()
            return

        self._launchFitBtn.hide()
        self.sigFitClicked.emit()
        # TODO: Abort button is not working
        # self._abortFit.show()

    def _computeResiduals(self) -> numpy.ndarray | None:
        """Note: The computation is cached as long as the dataset is loaded."""
        if self.dataset is None:
            missing_dataset_msg()
            return

        if self._residuals_cache is not None:
            return self._residuals_cache

        self._residuals_cache = numpy.sqrt(
            numpy.subtract(
                self._update_dataset.zsum(self.indices), self.dataset.zsum(self.indices)
            )
            ** 2
        )
        return self._residuals_cache

    def _displayMap(self, map_name: str):
        """
        :param map_name: Name of the map to display.
        """
        if self.dataset is None:
            return

        if self.maps is None:
            return

        title = self.dataset.title
        if title:
            graph_title = f"{title} - {map_name}"
        else:
            graph_title = map_name

        self._plotMaps.setKeepDataAspectRatio(True)
        self._plotMaps.setGraphTitle(graph_title)
        if map_name == "Residuals":
            self._addImage(self._computeResiduals())
            return

        self._addImage(self.maps[self._mapsCB.currentIndex()])

    def __abort(self):
        self._abortFitBtn.setEnabled(False)
        if self.dataset is not None:
            self.dataset.stop_operation(Operation.FIT)

    def onFitFinished(self):
        self._abortFitBtn.hide()
        self._launchFitBtn.show()

    def updateDataset(self, dataset: dtypes.ImageDataset, maps: numpy.ndarray):
        self._update_dataset, self.maps = dataset, maps
        self._displayMap(self._mapsCB.currentText())
        self._plotMaps.show()
        self._mapsCB.show()
        self._exportButton.show()

    def getIntensityThreshold(self) -> str:
        return self._intensityThresholdLE.text()

    def setIntensityThreshold(self, value: str):
        self._intensityThresholdLE.setText(value)

    def getFitMethod(self) -> str:
        return self._fitMethodCB.currentText()

    def setFitMethod(self, value: str):
        self._fitMethodCB.setCurrentText(value)

    def exportMaps(self):
        """
        Creates dictionary with maps information and exports it to a nexus file
        """
        if self.dataset is None or self.maps is None:
            missing_dataset_msg()
            return

        filename = select_output_hdf5_file_with_dialog()
        if filename:
            nxdict = generate_rocking_curves_nxdict(
                dataset=self.dataset, maps=self.maps, residuals=self._computeResiduals()
            )
            dicttonx(nxdict, filename)

    def _addImage(self, image):
        if self.dataset.transformation is None:
            self._plotMaps.addImage(image, xlabel="pixels", ylabel="pixels")
            return
        if self.dataset.transformation.rotate:
            image = numpy.rot90(image, 3)
        self._plotMaps.addImage(
            image,
            origin=self.dataset.transformation.origin,
            scale=self.dataset.transformation.scale,
            xlabel=self.dataset.transformation.label,
            ylabel=self.dataset.transformation.label,
        )



__authors__ = ["J. Garriga"]
__license__ = "MIT"
__date__ = "26/04/2021"

import logging

import silx
from packaging.version import Version
from silx.gui import qt
from silx.gui.colors import Colormap
from silx.gui.plot.items.roi import RectangleROI
from silx.gui.plot.StackView import StackViewMainWindow
from silx.gui.plot.tools.roi import RegionOfInterestManager

import darfix
from darfix import dtypes
from darfix.core.roi import clampROI

from .roiLimitsToolbar import RoiLimitsToolBar

_logger = logging.getLogger(__file__)


class ROISelectionWidget(qt.QWidget):
    """
    Widget that allows the user to pick a ROI in any image of the dataset.
    """

    sigValidate = qt.Signal()
    """Emit when user wants to keep the current resized dataset and move to next widget"""

    sigApply = qt.Signal()
    """Emit when user wants to Apply a ROI. Then we wait for a refinement or a validation"""

    sigReset = qt.Signal()
    """Emit when user wants to come back to the initial dataset"""

    sigROIChanged = qt.Signal()
    """Emit when the roi is changed"""

    def __init__(self, parent=None):
        super().__init__(parent)

        self.setLayout(qt.QVBoxLayout())
        self._sv = StackViewMainWindow()
        _buttons = qt.QDialogButtonBox(parent=self)

        self._okB = _buttons.addButton(_buttons.Ok)
        self._applyB = _buttons.addButton(_buttons.Apply)
        self._resetB = _buttons.addButton(_buttons.Reset)
        self._okB.setDefault(False)

        self._sv.setColormap(
            Colormap(
                name=darfix.config.DEFAULT_COLORMAP_NAME,
                normalization=darfix.config.DEFAULT_COLORMAP_NORM,
            )
        )
        self._sv.setKeepDataAspectRatio(True)
        self.layout().addWidget(self._sv)
        self.layout().addWidget(_buttons)

        if Version(silx.version) < Version("2.0.0"):
            plot = self._sv.getPlot()
        else:
            plot = self._sv.getPlotWidget()

        self._roiManager = RegionOfInterestManager(plot)

        self._roi = RectangleROI()
        if Version(silx.version) < Version("2.0.0"):
            self._roi.setLabel("ROI")
        else:
            self._roi.setText("ROI")
        self._roi.setGeometry(origin=(0, 0), size=(10, 10))
        self._roi.setEditable(True)
        self._roiManager.addRoi(self._roi)

        self._roiToolBar = RoiLimitsToolBar(roiManager=self._roiManager)
        self._sv.addToolBar(qt.Qt.BottomToolBarArea, self._roiToolBar)

        # connect signal / slot
        self._applyB.clicked.connect(self.applyRoi)
        self._okB.clicked.connect(self.sigValidate)
        self._resetB.clicked.connect(self.sigReset)
        # self._roiManager.sigRoiChanged.connect(self.sigROIChanged)
        self._roi.sigRegionChanged.connect(self.sigROIChanged)

    def setDataset(self, dataset: dtypes.Dataset):
        """Saves the dataset and updates the stack with the dataset data."""
        if dataset.dataset.title != "":
            self._sv.setGraphTitle(dataset.dataset.title)
        self.setStack(dataset=dataset.dataset)

    def _setROIForNewDataset(self, dataset: dtypes.ImageDataset):
        if not isinstance(dataset, dtypes.ImageDataset):
            raise dtypes.DatasetTypeError(dataset)
        first_frame_shape = dataset.frame_shape
        center = first_frame_shape[1] // 2, first_frame_shape[0] // 2
        size = first_frame_shape[1] // 5, first_frame_shape[0] // 5
        self.setRoi(center=center, size=size)

    def setStack(self, dataset: dtypes.ImageDataset):
        """
        Sets new data to the stack.
        Maintains the current frame showed in the view.

        :param Dataset dataset: if not None, data set to the stack will be from the given dataset.
        """
        if not isinstance(dataset, dtypes.ImageDataset):
            raise TypeError(
                f"dataset is expected to be an instance of {dtypes.ImageDataset}. Got {type(dataset)}."
            )

        nframe = self._sv.getFrameNumber()
        self._sv.setStack(dataset.as_array3d())
        self._sv.setFrameNumber(nframe)

    def setRoi(self, roi=None, origin=None, size=None, center=None):
        """
        Sets a region of interest of the stack of images.

        :param RectangleROI roi: A region of interest.
        :param Tuple origin: If a roi is not provided, used as an origin for the roi
        :param Tuple size: If a roi is not provided, used as a size for the roi.
        :param Tuple center: If a roi is not provided, used as a center for the roi.
        """
        if roi is not None and (
            size is not None or center is not None or origin is not None
        ):
            _logger.warning(
                "Only using provided roi, the rest of parameters are omitted"
            )

        if roi is not None:
            self._roi = roi
        else:
            self._roi.setGeometry(origin=origin, size=size, center=center)

    def getRoi(self):
        """
        Returns the roi selected in the stackview.

        :rtype: silx.gui.plot.items.roi.RectangleROI
        """
        return self._roi

    def applyRoi(self):
        """
        Function to apply the region of interest at the data of the dataset
        and show the new data in the stack. Dataset data is not yet replaced.
        A new roi is created in the middle of the new stack.
        """
        if self._sv.getPlotWidget().getActiveImage() is None:
            return

        self.enableApplyingROI(False)
        self.sigApply.emit()

    def enableApplyingROI(self, enable: bool) -> None:
        self._applyB.setEnabled(enable)

    def _fitROIToDataset(self, dataset: dtypes.ImageDataset):
        """
        This function is called when the ROI is applied to the dataset.
        The ROI must be updated to match the contour of the new shape.
        """
        dataset_shape = dataset.frame_shape
        self._roi.setGeometry(
            size=dataset_shape[::-1],
            center=(dataset_shape[1] / 2.0, dataset_shape[0] / 2.0),
        )

    def setVisible(self, visible):
        super().setVisible(visible)  # sets okB as default
        self._okB.setDefault(False)

    def getDataset(self) -> dtypes.Dataset:
        return self._update_dataset

    def getStackViewColormap(self):
        """
        Returns the colormap from the stackView

        :rtype: silx.gui.colors.Colormap
        """
        return self._sv.getColormap()

    def setStackViewColormap(self, colormap):
        """
        Sets the stackView colormap

        :param colormap: Colormap to set
        :type colormap: silx.gui.colors.Colormap
        """
        self._sv.setColormap(colormap)

    def clearStack(self):
        """
        Clears stack.
        """
        self._sv.setStack(None)
        self._roi.setGeometry(origin=(0, 0), size=(10, 10))

    def clampRoiToDataset(self, dataset: dtypes.ImageDataset):
        frame_height, frame_width = dataset.frame_shape
        # warning: we need to invert the order of the frame shape (frame_height, frame_width) vs (frame_width, frame_height)
        new_origin, new_size = clampROI(
            roi_origin=self._roi.getOrigin(),
            roi_size=self._roi.getSize(),
            frame_origin=(0, 0),
            frame_size=(frame_width, frame_height),
        )
        self._roi.setGeometry(
            origin=new_origin,
            size=new_size,
        )

    def enableValidation(self, isAllowed: bool) -> None:
        self._okB.setEnabled(isAllowed)



import numpy
import pytest

from darfix import dtypes
from darfix.tasks.datapartition import DataPartition
from darfix.tests.utils import createDataset


@pytest.fixture
def one_motor_dataset():
    # Dataset with scan shape 100, and frame shape 3,3, with frame intensity gradually increasing from 1 to 10
    return dtypes.Dataset(
        dataset=createDataset(
            data=numpy.linspace(1, 10, 100).repeat(9).reshape((100, 3, 3))
        ),
    )


def test_data_partition(one_motor_dataset):

    # test default processing
    task = DataPartition(
        inputs={
            "dataset": one_motor_dataset,
        }
    )
    task.run()

    # if no filtering then indices of all frames must exist
    numpy.testing.assert_array_equal(
        task.outputs.dataset.indices,
        numpy.arange(0, 100, step=1),
    )


def test_data_partition_with_filtering(one_motor_dataset):

    # test filtering
    task = DataPartition(
        inputs={
            "dataset": one_motor_dataset,
            "filter_bottom_bin_idx": 5,
            "filter_top_bin_idx": 45,
        },
    )
    task.run()

    # if filtering then some indices must be ignored
    assert not numpy.array_equal(
        task.outputs.dataset.indices,
        numpy.arange(0, 100, step=1),
    )



from __future__ import annotations

import h5py
from silx.gui import qt

from darfix.core.datapathfinder import DETECTOR_KEYWORD
from darfix.core.datapathfinder import FIRST_SCAN_KEYWORD
from darfix.core.datapathfinder import LAST_SCAN_KEYWORD
from darfix.gui.utils.data_path_selection import DataPathSelection
from darfix.gui.utils.fileselection import FileSelector
from darfix.tasks.hdf5_scans_concatenation import ConcatenateHDF5Scans


class HDF5DatasetSelectionWidget(qt.QWidget):
    """Widget to concatenate a series of scans together"""

    sigInputFileChanged = qt.Signal(str)

    DETECTOR_PATH_ALLOWED_KEYWORDS = (
        DETECTOR_KEYWORD,
        FIRST_SCAN_KEYWORD,
        LAST_SCAN_KEYWORD,
    )

    POSITIONER_PATH_ALLOWED_KEYWORDS = (
        FIRST_SCAN_KEYWORD,
        LAST_SCAN_KEYWORD,
    )

    def __init__(
        self,
        parent: qt.QWidget | None = None,
    ) -> None:
        super().__init__(parent)
        self.setLayout(qt.QGridLayout())
        # input file selector
        self._inputFileSelector = FileSelector(label="input file:")
        self.layout().addWidget(self._inputFileSelector, 0, 0, 1, 5)
        self._inputFileSelector.setDialogFileMode(qt.QFileDialog.ExistingFile)
        self._inputFileSelector.setDialogNameFilters(
            ("HDF5 files (*.h5 *.hdf5 *.nx *.nxs *.nexus)",)
        )

        # detector data path
        self._detectorDataPath = DataPathSelection(
            self,
            title="detector-data path",
            completer_display_dataset=True,
            data_path_type=h5py.Dataset,
        )
        self._detectorDataPath.setPattern(
            ConcatenateHDF5Scans.DEFAULT_DETECTOR_DATA_PATH,
            store_as_default=True,
        )
        self._detectorDataPath.setAvailableKeywords(self.DETECTOR_PATH_ALLOWED_KEYWORDS)
        self.layout().addWidget(self._detectorDataPath, 1, 0, 1, 5)
        # positioner data path
        self._positionerDataPath = DataPathSelection(
            self,
            title="metadata path (positioners)",
            completer_display_dataset=False,
            data_path_type=h5py.Group,
        )
        self._positionerDataPath.setPattern(
            ConcatenateHDF5Scans.DEFAULT_POSITIONERS_DATA_PATH,
            store_as_default=True,
        )
        self._positionerDataPath.setAvailableKeywords(
            self.POSITIONER_PATH_ALLOWED_KEYWORDS
        )
        self.layout().addWidget(self._positionerDataPath, 2, 0, 1, 5)

        # spacer
        spacer = qt.QWidget()
        spacer.setSizePolicy(qt.QSizePolicy.Expanding, qt.QSizePolicy.Expanding)
        self.layout().addWidget(spacer, 999, 0, 1, 5)

        # connect signal / slot
        self._inputFileSelector.sigFileChanged.connect(self._inputFileChanged)

    def _inputFileChanged(self):
        input_file = self.getInputFile()
        self._detectorDataPath.setInputFile(input_file)
        self._positionerDataPath.setInputFile(input_file)
        self.sigInputFileChanged.emit(input_file)

    def getDetectorDataPathSelection(self) -> DataPathSelection:
        return self._detectorDataPath

    def getMetadataPathSelection(self) -> DataPathSelection | None:
        return self._positionerDataPath

    def setInputFile(self, file_path: str):
        self._inputFileSelector.setFilePath(file_path=file_path)
        self._inputFileChanged()

    def getInputFile(self) -> str | None:
        return self._inputFileSelector.getFilePath()


class HDF5RawDatasetSelectionWidget(HDF5DatasetSelectionWidget):
    """Same as the 'HDF5DatasetSelectionWidget' but adds an option to keep data on disk and to provide the title"""

    sigKeepDataOnDiskChanged = qt.Signal(bool)
    sigTitleChanged = qt.Signal(str)

    def __init__(self, parent: qt.QWidget | None = None) -> None:
        super().__init__(parent)
        self._keepDataOnDisk = qt.QCheckBox("Keep data on disk")
        self.layout().addWidget(self._keepDataOnDisk, 1, 3, 2, 2)

        self._titleLabel = qt.QLabel("Workflow title")
        self.layout().addWidget(self._titleLabel, 3, 0, 1, 1)
        self._title = qt.QLineEdit("")
        self.layout().addWidget(self._title, 3, 1, 1, 4)

        # redefine layout because we want to display 'title' just after the file path selection
        self.layout().addWidget(self._detectorDataPath, 4, 0, 1, 5)
        self.layout().addWidget(self._positionerDataPath, 5, 0, 1, 5)

        # connect signal / slot
        self._keepDataOnDisk.toggled.connect(self.sigKeepDataOnDiskChanged)
        self._title.editingFinished.connect(self._titleChanged)

    def isKeepingDataOnDisk(self) -> bool:
        return self._keepDataOnDisk.isChecked()

    def setKeepDataOnDisk(self, value: bool):
        self._keepDataOnDisk.setChecked(value)

    def getWorkflowTitle(self) -> str:
        return self._title.text()

    def setWorkflowTitle(self, title: str):
        self._title.setText(title)
        self._titleChanged()

    def _titleChanged(self, *args, **kwargs):
        self.sigTitleChanged.emit(self._title.text())


class HDF5DarkDatasetSelectionWidget(HDF5DatasetSelectionWidget):
    """Same as the 'HDF5DatasetSelectionWidget' but hide metadata selection that is unused for dark"""

    def __init__(self, parent: qt.QWidget | None = None) -> None:
        super().__init__(parent)
        # no positioner possible for dark so hide the selection
        self._positionerDataPath.hide()

    def getMetadataPathSelection(self) -> None:
        return None



from __future__ import annotations

import logging
import threading
from typing import Any

from ewokscore import TaskWithProgress
from ewokscore.missing_data import MISSING_DATA
from ewokscore.missing_data import MissingData
from ewokscore.model import BaseInputModel
from pydantic import ConfigDict
from pydantic import Field

from darfix.core.state_of_operation import Operation
from darfix.dtypes import Dataset

from ..core.noiseremoval import NoiseRemovalOperation
from ..core.noiseremoval import apply_noise_removal_operations
from ..core.noiseremoval import create_background_substraction_operation

_logger = logging.getLogger(__name__)


class Inputs(BaseInputModel):
    model_config = ConfigDict(use_attribute_docstrings=True)
    dataset: Dataset
    """ Input dataset containing a stack of images """
    operations: list[dict[str, Any]] | MissingData = Field(
        default=MISSING_DATA,
        examples=[
            [
                {"type": "THRESHOLD", "parameters": {"bottom": 10.0, "top": 1000.0}},
                {"type": "HP", "parameters": {"kernel_size": 3}},
            ]
        ],
        description="""List of noise removal operations to apply to the dataset. Empty list if not provided."

        Available operations :

        - 'Operation.THRESHOLD': Threshold operation. Parameters: 'bottom' (float) and 'top' (float). Keep value only if it is between bottom and top.
        - 'Operation.HP': Hot Pixel removal using median filter operation. Parameters: 'kernel_size' (int).
        - 'Operation.BS': Background subtraction operation. Parameters: 'method' ("mean" | "median") and 'background_type' ("Data" | "Unused data (after partition)" | "Dark data").
        - 'Operation.MASK': Mask removal operation. Parameters: 'mask' (numpy.ndarray 2D containing 0 and 1 where 0 indicates the pixels to be removed).
        """,
    )


class NoiseRemoval(
    TaskWithProgress,
    input_model=Inputs,
    output_names=["dataset"],
):
    """Apply a list of noise removal operations on a Darfix dataset."""

    def run(self):

        self.cancelEvent = threading.Event()

        input_dataset: Dataset = self.get_input_value("dataset")

        operations = []

        for operation in self.get_input_value("operations", []):
            operation = NoiseRemovalOperation(operation)
            if operation["type"] is Operation.BS:
                _logger.info("Computing background...")
                background_substraction_operation = (
                    create_background_substraction_operation(
                        input_dataset, **operation["parameters"]
                    )
                )
                operations.append(background_substraction_operation)
            else:
                operations.append(operation)

        apply_noise_removal_operations(
            input_dataset.dataset.as_array3d(input_dataset.indices),
            operations,
            self._is_cancelled,
            self._set_progress,
        )

        self.outputs.dataset = input_dataset

    def cancel(self) -> None:
        self.cancelEvent.set()

    def _is_cancelled(self) -> bool:
        return self.cancelEvent.is_set()

    def _set_progress(self, progress: int):
        self.progress = progress



import copy
import os
import tempfile
import unittest
import uuid

import numpy

from darfix.tests import utils


class _BaseDatasetTest:
    """Tests for class Dataset in `dataset.py`"""

    def createRandomDataset(self, dims, nb_frames, metadata=False):
        raise NotImplementedError("Base class")

    def test_data_load(self):
        """Tests the correct load of the data"""
        self.assertEqual(len(self.dataset.as_array3d()), 3)
        self.assertEqual(self.dataset.nframes, 3)

    def test_nframes(self):
        """Tests the nframes method"""
        self.assertEqual(self.dataset.nframes, 3)

    def test_deepcopy(self):
        """Tests the correct deepcopy of the object"""
        dataset_copy = copy.deepcopy(self.dataset)
        self.assertEqual(self.dataset.nframes, dataset_copy.nframes)
        self.assertEqual(self.dataset.data.shape, dataset_copy.data.shape)

    def test_zsum(self):
        zsum = self.dataset.zsum()
        self.assertEqual(zsum.shape, (100, 100))

        result = numpy.sum(self.dataset.as_array3d(), axis=0)
        numpy.testing.assert_array_equal(zsum, result)

        indices = [0, 1]
        zsum = self.dataset.zsum(indices)
        result = numpy.sum(self.dataset.as_array3d(indices), axis=0)
        numpy.testing.assert_array_equal(zsum, result)

    def test_filter_data(self):
        """Tests the correct separation of empty frames and data frames"""
        dims = (10, 100, 100)
        data = numpy.zeros(dims)
        background = numpy.random.random(dims)
        idxs = [0, 2, 4]
        data[idxs] += background[idxs]
        dataset = utils.createDataset(data=data, _dir=self._dir)
        used_idx, not_used_idx = dataset.partition_by_intensity(bottom_bin=1)
        self.assertEqual(used_idx.shape[0], 3)
        self.assertEqual(not_used_idx.shape[0], 7)

    def test_roi(self):
        """Tests the roi function"""
        new_dataset = self.dataset.apply_roi(origin=[0, 0], size=[20, 20])
        self.assertEqual(new_dataset.nframes, 3)
        numpy.testing.assert_equal(self.dataset.data[:, :20, :20], new_dataset.data)

    def test_find_shift(self):
        """Tests the shift detection"""
        shift = self.dataset.find_shift()

        self.assertEqual(len(shift), 2)

        shift = self.dataset.find_shift(indices=[0, 1])

        self.assertEqual(len(shift), 2)

    def test_pca(self):

        H, W = self.dataset.pca()

        self.assertEqual(
            H.shape, (self.dataset.nframes, len(self.dataset.data[0].flatten()))
        )
        self.assertEqual(W.shape, (self.dataset.nframes, self.dataset.nframes))

        n_components = 1
        indices = [1, 2]
        H, W = self.dataset.pca(num_components=n_components, indices=indices)

        self.assertEqual(H.shape, (n_components, len(self.dataset.data[0].flatten())))
        self.assertEqual(W.shape, (2, n_components))

    def test_nmf(self):

        num_components = 2
        H, W = self.dataset.nmf(num_components)

        self.assertEqual(H.shape, (num_components, len(self.dataset.data[0].flatten())))
        self.assertEqual(W.shape, (self.dataset.nframes, num_components))

        num_components = 1
        indices = [1, 2]
        H, W = self.dataset.nmf(num_components=num_components, indices=indices)

        self.assertEqual(H.shape, (num_components, len(self.dataset.data[0].flatten())))
        self.assertEqual(W.shape, (2, num_components))

    def test_nica(self):

        num_components = 2
        H, W = self.dataset.nica(num_components)

        self.assertEqual(H.shape, (num_components, len(self.dataset.data[0].flatten())))
        self.assertEqual(W.shape, (self.dataset.nframes, num_components))

        num_components = 1
        indices = [1, 2]
        H, W = self.dataset.nica(num_components=num_components, indices=indices)

        self.assertEqual(H.shape, (num_components, len(self.dataset.data[0].flatten())))
        self.assertEqual(W.shape, (2, num_components))


class TestHDF5Dataset(_BaseDatasetTest, unittest.TestCase):
    def setUp(self):
        self._dir = tempfile.mkdtemp()
        output_file = os.path.join(self._dir, str(uuid.uuid1()) + ".hdf5")
        self.dataset = utils.createRandomHDF5Dataset(
            dims=(100, 100), nb_data_frames=3, metadata=True, output_file=output_file
        )

    def createRandomDataset(self, dims, nb_frames, metadata=False):
        return utils.createRandomHDF5Dataset(
            dims=dims, nb_data_frames=nb_frames, metadata=metadata
        )


def test_reshaped_data(dataset):
    """Tests the correct reshaping of the data"""

    dataset.find_dimensions()
    dataset.reshape_data()
    assert dataset.data.shape == (2, 2, 5, 100, 100)


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

from typing import Literal

from ewokscore import Task
from ewokscore.missing_data import MISSING_DATA
from ewokscore.missing_data import MissingData
from ewokscore.model import BaseInputModel
from pydantic import ConfigDict

from ..dtypes import Dataset
from ..math import Vector3D
from ..pixel_sizes import PixelSize


class Inputs(BaseInputModel):
    model_config = ConfigDict(use_attribute_docstrings=True)
    dataset: Dataset
    """Input dataset containing a stack of images."""
    Q: Vector3D
    """Scattering vector in oriented pseudocubic coordinates."""
    a: float
    """Pseudocubic lattice parameter."""
    map_range: float
    """Range (in all 3 directions) of the histogram. Center-to-edge distance."""
    detector: str
    """Detector type for the RSM computation."""
    units: Literal["poulsen", "gorfman"] | MissingData = MISSING_DATA
    """Either 'poulsen' [10.1107/S1600576717011037] or 'gorfman' [https://arxiv.org/pdf/2110.14311.pdf]. 'poulsen' if not provided"""
    n: Vector3D | MissingData = MISSING_DATA
    """Surface normal of the sample in oriented pseudocubic hkl"""
    map_shape: Vector3D | MissingData = MISSING_DATA
    """Number of bins in each direction."""
    energy: float | MissingData = MISSING_DATA


class RSMHistogram(
    Task,
    input_model=Inputs,
    output_names=["hist_values", "hist_edges"],
):
    """Computes Reciprocal Space Map histogram."""

    def run(self):
        input_dataset = self.inputs.dataset
        if not isinstance(input_dataset, Dataset):
            raise TypeError(
                f"dataset is expected to be an instance of Dataset. Got {type(input_dataset)}."
            )
        dataset = input_dataset.dataset

        units: str | None = self.get_input_value("units", None)
        n: Vector3D | None = self.get_input_value("n", None)
        map_shape: Vector3D | None = self.get_input_value("map_shape", None)
        energy: float | None = self.get_input_value("energy", None)

        values, edges = dataset.compute_rsm(
            Q=self.inputs.Q,
            a=self.inputs.a,
            map_range=self.inputs.map_range,
            pixel_size=PixelSize[self.inputs.detector].value,
            units=units.lower() if units else None,
            n=n,
            map_shape=map_shape,
            energy=energy,
        )

        self.outputs.hist_values = values
        self.outputs.hist_edges = edges



from __future__ import annotations

from typing import Sequence

from ewokscore import Task
from ewokscore.model import BaseInputModel
from pydantic import ConfigDict

from darfix.core.shiftcorrection import apply_shift
from darfix.dtypes import Dataset


class Inputs(BaseInputModel):
    model_config = ConfigDict(use_attribute_docstrings=True)
    dataset: Dataset
    """ Input dataset containing a stack of images """
    shift: Sequence[float] | Sequence[Sequence[float]] | None = None
    """Shift to apply to the images. If not provided, dataset will be unchanged."""
    selected_axis: int | None = None
    """Selected dimension axis. If not None. We considere a linear shift along this dimension.  Darfix convention is : dimension with axis 0 is the fast motor."""


class ShiftCorrection(
    Task,
    input_model=Inputs,
    output_names=["dataset"],
):
    def run(self):
        inputs = Inputs(**self.get_input_values())

        if inputs.shift is None:
            self.outputs.dataset = inputs.dataset
            return

        apply_shift(inputs.dataset, inputs.shift, inputs.selected_axis)

        self.outputs.dataset = inputs.dataset



from __future__ import annotations

from typing import Any
from typing import Dict

from ewoksorange.bindings.owwidgets import OWEwoksWidgetOneThread
from ewoksorange.gui.parameterform import block_signals
from silx.gui import qt

from darfix import dtypes
from darfix.gui.magnificationWidget import MagnificationWidget
from darfix.gui.rsmWidget import RSMWidget
from darfix.tasks.transformation import TransformationMatrixComputation


class TransformationWidgetOW(
    OWEwoksWidgetOneThread, ewokstaskclass=TransformationMatrixComputation
):
    """
    Widget to computes transformation matrix.
    """

    name = "transformation"
    icon = "icons/axes.png"
    want_main_area = False

    _ewoks_inputs_to_hide_from_orange = (
        "kind",
        "orientation",
        "magnification",
        "pixelSize",
        "rotate",
    )

    def __init__(self):
        super().__init__()
        self._magWidget = MagnificationWidget(self)
        magnification = self.get_default_input_value("magnification", None)
        if magnification:
            self._magWidget.magnification = magnification
        orientation = self.get_default_input_value("orientation", None)
        if orientation:
            self._magWidget.orientation = orientation

        self._rsmWidget = RSMWidget(self)
        pixelSize = self.get_default_input_value("pixelSize", None)
        if pixelSize:
            self._rsmWidget.pixelSize = pixelSize
        rotate = self.get_default_input_value("rotate", None)
        if rotate:
            self._rsmWidget.rotate = rotate
        self._stackedWidget = qt.QStackedWidget(self)
        self._stackedWidget.addWidget(self._magWidget)
        self._stackedWidget.addWidget(self._rsmWidget)

        self._methodCB = qt.QComboBox(self)
        self._methodCB.currentTextChanged.connect(self._changeTransformationWidget)

        self._okButton = qt.QPushButton(self, text="Ok")
        self._okButton.clicked.connect(self._execute_task)
        self._okButton.setEnabled(False)

        layout = self._get_control_layout()
        layout.addWidget(self._methodCB)
        layout.addWidget(self._stackedWidget)
        layout.addWidget(self._okButton)

    def handleNewSignals(self):
        dataset = self.get_task_input_value("dataset", None)
        if dataset is not None:
            self.setDataset(dataset)

        # Do not call super().handleNewSignals() to prevent propagation

    def setDataset(self, _input: dtypes.Dataset):
        if not isinstance(_input, dtypes.Dataset):
            raise TypeError(
                f"_input is expected to be an instance of {dtypes.Dataset}. Got {type(_input)}."
            )

        darfix_dataset = _input.dataset

        ndim = darfix_dataset.dims.ndim
        if ndim == 0:
            msg = qt.QMessageBox()
            msg.setIcon(qt.QMessageBox.Warning)
            msg.setText(
                "This widget has to be used before selecting any region of \
                            interest and after selecting the dimensions"
            )
            msg.exec()
            return

        self._dataset = _input
        self._changeDimensions(ndim)
        self._okButton.setEnabled(True)

    def _changeDimensions(self, ndim: int) -> None:
        """The possible transformations depend on the number of dimensions (one or two)."""
        if ndim == 1:
            transformations = ["Magnification", "RSM"]
        elif ndim == 2:
            transformations = ["Magnification"]
        else:
            raise ValueError(
                f"The Transformation Widget only works with 1D or 2D datasets. Not {ndim}D."
            )

        current_items = [
            self._methodCB.itemText(i) for i in range(self._methodCB.count())
        ]
        if transformations == current_items:
            return

        with block_signals(self._methodCB):
            old_method = self._methodCB.currentText()
            self._methodCB.clear()
            self._methodCB.addItems(transformations)
            try:
                idx = transformations.index(old_method)
            except ValueError:
                idx = 0
            self._methodCB.setCurrentIndex(idx)

    def _changeTransformationWidget(self, method: str) -> None:
        """
        Change the widget displayed on the window
        """

        if method == "RSM":
            self._stackedWidget.setCurrentWidget(self._rsmWidget)
        elif method == "Magnification":
            self._stackedWidget.setCurrentWidget(self._magWidget)
        else:
            return

    def _execute_task(self) -> None:
        method = self._methodCB.currentText()

        args: Dict[str, Any] = {"dataset": self._dataset}
        if method == "Magnification":
            args["magnification"] = self._magWidget.magnification
            args["kind"] = "magnification"
            args["orientation"] = self._magWidget.orientation
        elif method == "RSM":
            args["pixelSize"] = self._rsmWidget.pixelSize
            args["rotate"] = self._rsmWidget.rotate
            args["kind"] = "rsm"
        self.close()

        for name, value in args.items():
            self.set_default_input(name, value)

        self.execute_ewoks_task()



from __future__ import annotations

from numbers import Number

import numpy


class Transformation:
    def __init__(self, kind: str, x: numpy.ndarray, y: numpy.ndarray, rotate: bool):
        self.x = x  # shape (sy, sx)
        self.y = y  # shape (sy, sx)
        self.rotate = rotate
        self.kind = kind

    @property
    def shape(self) -> tuple[int, int]:
        return self.x.shape

    @property
    def sx(self) -> int:
        return self.shape[1]

    @property
    def sy(self) -> int:
        return self.shape[0]

    @property
    def label(self) -> str:
        return "degrees" if self.kind == "rsm" else "µm"

    @property
    def xregular(self) -> numpy.ndarray:
        return self.xscale * numpy.arange(self.sx) + self.xorigin

    @property
    def yregular(self) -> numpy.ndarray:
        return self.yscale * numpy.arange(self.sy) + self.yorigin

    @property
    def xorigin(self) -> Number:
        return self.x[0][0]

    @property
    def yorigin(self) -> Number:
        return self.y[0][0]

    @property
    def origin(self) -> tuple[Number, Number]:
        return self.xorigin, self.yorigin

    @property
    def xscale(self) -> Number:
        x = self.x
        return (x[-1][-1] - x[0][0]) / self.sx

    @property
    def yscale(self) -> Number:
        y = self.y
        return (y[-1][-1] - y[0][0]) / self.sy

    @property
    def scale(self) -> tuple[Number, Number]:
        return self.xscale, self.yscale



from __future__ import annotations

from typing import Literal
from typing import Tuple

import numpy
import tqdm
from numba import njit
from numba import prange
from silx.math.medianfilter import medfilt2d
from skimage.transform import rescale

import darfix

from ..math import Vector3D


@njit(parallel=True, fastmath=True)
def _mean_fast(
    x: numpy.ndarray,
    y: numpy.ndarray,
    zsum: numpy.ndarray,
    n_frames: int,
    height: int,
    width: int,
) -> numpy.ndarray:

    mean = numpy.empty((height, width), dtype=numpy.float64)
    for j in prange(height):
        for i in range(width):
            if zsum[j, i] == 0.0:
                mean[j, i] = numpy.nan
                continue
            weighted_sum = 0.0
            for k in range(n_frames):
                fk = y[k, j, i]
                weighted_sum += fk * x[k]
            mean[j, i] = weighted_sum / zsum[j, i]
    return mean


@njit(parallel=True, fastmath=True)
def _sigma_fast(
    x: numpy.ndarray,
    y: numpy.ndarray,
    zsum: numpy.ndarray,
    mean: numpy.ndarray,
    n_frames: int,
    height: int,
    width: int,
) -> numpy.ndarray:

    sigma = numpy.empty((height, width), dtype=numpy.float64)

    for j in prange(height):
        for i in range(width):
            if zsum[j, i] == 0.0:
                sigma[j, i] = numpy.nan
                continue

            _sum = 0
            for k in range(n_frames):
                _sum += (x[k] - mean[j, i]) ** 2 * y[k, j, i]

            sigma[j, i] = numpy.sqrt(_sum / zsum[j, i])
    return sigma


@njit(parallel=True, fastmath=True)
def _skewness_kurtosis_fast(
    x: numpy.ndarray,
    y: numpy.ndarray,
    zsum: numpy.ndarray,
    mean: numpy.ndarray,
    sigma: numpy.ndarray,
    n_frames: int,
    height: int,
    width: int,
) -> tuple[numpy.ndarray, numpy.ndarray]:

    skew = numpy.empty((height, width), dtype=numpy.float64)
    kurt = numpy.empty((height, width), dtype=numpy.float64)
    for j in prange(height):
        for i in range(width):
            if zsum[j, i] == 0.0 or sigma[j, i] == 0.0:
                skew[j, i] = numpy.nan
                kurt[j, i] = numpy.nan
                continue

            _sum3 = 0.0
            _sum4 = 0.0

            for k in range(n_frames):
                z = (x[k] - mean[j, i]) / sigma[j, i]
                skew_k = z**3 * y[k, j, i]
                _sum3 += skew_k
                _sum4 += skew_k * z

            skew[j, i] = _sum3 / zsum[j, i]
            kurt[j, i] = (_sum4 / zsum[j, i]) - 3.0

    return skew, kurt


def compute_moments(motor_values, data, smooth: bool = True):
    """
    Compute first, second, third and fourth moment of data on values.

    :param motor_values: 1D array of X-values
    :param data: nD array of Y-values with `len(data) == len(motor_values)`
    :param smooth: if True apply a 2D median filter on moments
    :returns: The four first moments to distribution Y(X)
    """
    if len(motor_values) != len(data):
        raise ValueError("the length of 'values' and 'data' is not equal")

    # Moments
    # mean = sum(y * x) / sum(y)
    # var  = sum(y * (x - mean)^2) / sum(y)
    # skew = sum(y * ((x - mean)/sigma)^3) / sum(y)
    # kurt = sum(y * ((x - mean)/sigma)^4) / sum(y) - 3
    n_frames, height, width = data.shape

    zsum = numpy.sum(data, axis=0)

    with tqdm.tqdm(total=3) as progressbar:
        progressbar.set_description("Compute COM...")
        mean = _mean_fast(motor_values, data, zsum, n_frames, height, width)
        progressbar.update()
        progressbar.set_description("Compute FWHM...")
        sigma = _sigma_fast(motor_values, data, zsum, mean, n_frames, height, width)
        fwhm = darfix.config.FWHM_VAL * sigma
        progressbar.update()
        progressbar.set_description("Compute Skewness and Kurtosis...")
        skew, kurt = _skewness_kurtosis_fast(
            motor_values, data, zsum, mean, sigma, n_frames, height, width
        )
        progressbar.update()

    if smooth:
        mean = medfilt2d(mean)
        fwhm = medfilt2d(fwhm)
        skew = medfilt2d(skew)
        kurt = medfilt2d(kurt)

    return mean, fwhm, skew, kurt


def compute_peak_position(data, values=None, center_data=False):
    """
    Compute peak position map

    :param bool center_data: If True, the values are centered on 0.
    """
    if values is not None:
        values = numpy.asanyarray(values)
        x = numpy.array(numpy.argmax(data, axis=0))
        if center_data:
            middle = float(min(values) + numpy.ptp(values)) / 2
            values -= middle
        image = [values[i] for i in x.flatten()]
        image = numpy.reshape(image, x.shape)
    else:
        image = numpy.array(numpy.argmax(data, axis=0))
        if center_data:
            middle = len(data) / 2
            vals = numpy.linspace(-middle, middle, len(data))
            image = image * numpy.ptp(vals) / len(data) + numpy.min(vals)
    return image


def compute_rsm(
    H: int, W: int, d: float, ffz: float, mainx: float, rotate: bool = False
) -> Tuple[numpy.ndarray, numpy.ndarray]:
    """
    Transformation to azimuthal coordinates.

    :param int H: height of the image in pixels.
    :param int W: width of the image in pixels.
    :param float d: Distance in micrometers of each pixel.
    :param float ffz: motor 'ffz' value.
    :param float mainx: motor 'mainx' value.

    :returns: Tuple of two arrays of size (W, H)
    :rtype: (X1, X2) : ndarray
    """
    if rotate:
        y = (numpy.arange(H) - H / 2) * d
        z = ffz - (W / 2 - numpy.arange(W)) * d
        y, z = numpy.meshgrid(y, z, indexing="ij")
    else:
        y = (numpy.arange(W) - W / 2) * d
        z = ffz - (H / 2 - numpy.arange(H)) * d
        z, y = numpy.meshgrid(z, y, indexing="ij")
    eta = numpy.arctan2(y, z)
    twotheta = numpy.arctan2(numpy.sqrt(y * y + z * z), mainx)
    return numpy.degrees(eta), numpy.degrees(twotheta)


def compute_magnification(
    H: int,
    W: int,
    d: float,
    obx: float,
    obpitch: float,
    mainx: float,
    topography_orientation: int | None = None,
    center: bool = True,
) -> Tuple[numpy.ndarray, numpy.ndarray]:
    """
    :param int H: height of the image in pixels.
    :param int W: width of the image in pixels.
    :param float d: Distance in micrometers of each pixel.
    :param float obx: motor 'obx' value.
    :param float obpitch: motor 'obpitch' value in the middle of the dataset.
    :param float mainx: motor 'mainx' value.

    :returns: Tuple of two arrays of size (H, W)
    :rtype: (X1, X2) : ndarray
    """

    pix_arr = list(numpy.meshgrid(numpy.arange(H), numpy.arange(W)))
    d1 = obx / numpy.cos(numpy.radians(obpitch))
    d2 = mainx / numpy.cos(numpy.radians(obpitch)) - d1
    M = d2 / d1
    d /= M
    if center:
        pix_arr[0] = (pix_arr[0] - W / 2) * d
        pix_arr[1] = (H / 2 - pix_arr[1]) * d
    else:
        pix_arr[0] = pix_arr[0] * d
        pix_arr[1] = (H - 1 - pix_arr[1]) * d
    if topography_orientation is not None:
        pix_arr[topography_orientation] /= numpy.sin(numpy.radians(obpitch))
    return pix_arr[0], pix_arr[1]


def rescale_data(data, scale):
    new_data = None
    for i, image in enumerate(data):
        simage = rescale(image, scale, anti_aliasing=True, preserve_range=True)
        if new_data is None:
            new_data = numpy.empty((len(data),) + simage.shape, dtype=data.dtype)
        new_data[i] = simage
    return new_data


def calculate_RSM_histogram(
    data: numpy.ndarray,
    diffry_values: numpy.ndarray,
    twotheta: numpy.ndarray,
    eta: numpy.ndarray,
    Q: Vector3D,
    a: float,
    map_range: float,
    units: Literal["poulsen", "gorfman"] | None = None,
    map_shape: Vector3D | None = None,
    n: Vector3D | None = None,
    E: float | None = None,
):
    """
    ***Code originally written by Mads Carslen***

    Calculate reciprocal space map from a 'diffry' scan without the objective lens.
    The RSM is calculated as a multidimensional histogram;

    :param data:
    :param diffry_values:
    :param twotheta:
    :param eta:
    :param Q: Scattering vector in oriented pseudocubic coordinates.
    :param a: pseudocubic lattice parameter
    :param map_range: range (in all 3 directions) of the histogram. Center-to edge-distance.
    :param units: either 'poulsen'  [10.1107/S1600576717011037] or 'gorfman' [https://arxiv.org/pdf/2110.14311.pdf]. Default: 'poulsen'
    :param map_shape: Number of bins in each direction
    :param n: surface normal of the sample in oriented pseudocubic hkl
    :param E: energy
    """

    if units is None:
        units = "poulsen"
    if map_shape is None:
        map_shape = (50, 50, 50)
    if n is None:
        n = (1, 0, 0)
    if E is None:
        E = 17.0

    k = 2 * numpy.pi / (12.391 / E)

    diffry_center = numpy.mean(diffry_values)
    img_shape = data[0].shape
    if units == "gorfman":
        # Build orientation matrix
        sampl_z = numpy.array(Q) / numpy.linalg.norm(
            numpy.array(Q)
        )  # assume scattering vector is z
        sampl_x = n - sampl_z * numpy.dot(sampl_z, n) / numpy.linalg.norm(
            n
        )  # orthogonalize
        sampl_x /= numpy.linalg.norm(sampl_x)  # nomalize

        sampl_y = numpy.cross(sampl_z, sampl_x)
        lab_to_lat = numpy.stack((sampl_x, sampl_y, sampl_z))
    elif units == "poulsen":
        lab_to_lat = numpy.identity(3)

    # Calculate lab frame q vector for each pixel
    k0 = numpy.array([k, 0, 0])  # Lab frame incidetn wavevector
    twotheta = numpy.radians(twotheta)
    eta = numpy.radians(eta)
    kh = k * numpy.stack(
        [
            numpy.cos(twotheta),
            numpy.sin(twotheta) * numpy.sin(eta),
            numpy.sin(twotheta) * numpy.cos(eta),
        ]
    )  # Lab frame scattered wavevector
    q = kh - k0[:, numpy.newaxis, numpy.newaxis]
    if units == "gorfman":
        q = q * a / 2 / numpy.pi
    elif units == "poulsen":
        q = q * a / 2 / numpy.pi / numpy.linalg.norm(Q)

    # flatten to match syntax for histogramdd
    q = q.reshape(3, img_shape[0] * img_shape[1])
    # Rotate from lab frame to sample frame
    theta_ref = numpy.arcsin(2 * numpy.pi * numpy.linalg.norm(Q) / a / k / 2)
    q = numpy.stack(
        [
            q[0, ...] * numpy.cos(theta_ref) + q[2, ...] * numpy.sin(theta_ref),
            q[1, ...],
            q[2, ...] * numpy.cos(theta_ref) - q[0, ...] * numpy.sin(theta_ref),
        ]
    )

    # Make histogram ranges
    q_mean = numpy.mean(q, axis=1)
    diffry_mean = numpy.radians(numpy.mean(diffry_values) - diffry_center)
    q_mean = numpy.stack(
        [
            q_mean[0] * numpy.cos(diffry_mean) - q_mean[2] * numpy.sin(diffry_mean),
            q_mean[1],
            q_mean[2] * numpy.cos(diffry_mean) + q_mean[0] * numpy.sin(diffry_mean),
        ]
    )

    if units == "gorfman":
        q_mean = lab_to_lat.transpose() @ q_mean

    ranges = (
        (q_mean[0] - map_range, q_mean[0] + map_range),
        (q_mean[1] - map_range, q_mean[1] + map_range),
        (q_mean[2] - map_range, q_mean[2] + map_range),
    )  # hkl units

    # initialize sum arrays
    sum_inte = numpy.zeros(map_shape)
    sum_freq = numpy.zeros(map_shape)

    loop = tqdm.tqdm(data)
    # Loop through images
    for i, image in enumerate(loop):
        # read angle
        diffry = numpy.radians(diffry_values[i] - diffry_center)
        # rotate q back to zero-motor frame
        q_rot = numpy.stack(
            [
                q[0, ...] * numpy.cos(diffry) - q[2, ...] * numpy.sin(diffry),
                q[1, ...],
                q[2, ...] * numpy.cos(diffry) + q[0, ...] * numpy.sin(diffry),
            ]
        )
        # Rotate into lattice frame
        q_rot = lab_to_lat.transpose() @ q_rot
        # Do binning
        sum_inte += numpy.histogramdd(
            q_rot.transpose(), map_shape, ranges, weights=image.flatten()
        )[0]
        sum_freq += numpy.histogramdd(q_rot.transpose(), map_shape, ranges)[0]

    edges = numpy.histogramdd(q_rot.transpose(), map_shape, ranges)[1]
    # Setting sum_freq to NaN where it is 0 to avoid division by 0 (`arr` will be NaN there)
    sum_freq[sum_freq == 0] = numpy.nan
    arr = sum_inte / sum_freq

    if units == "poulsen":
        edges[2] = edges[2] - 1

    return arr, edges



from __future__ import annotations

from ewoksorange.bindings.owwidgets import OWEwoksWidgetOneThread

from darfix import dtypes
from darfix.gui.grainplot.grainPlotWidget import GrainPlotWidget
from darfix.tasks.grainplot import GrainPlot


class GrainPlotWidgetOW(OWEwoksWidgetOneThread, ewokstaskclass=GrainPlot):
    """
    Computes moments (Center of mass, FWHM, Kurtosis, SKEWNESS) and displays them.

    Also computes mosaicity and orientation distribution for multi-dimensional datasets.
    """

    _ewoks_inputs_to_hide_from_orange = (
        "filename",
        "dimensions",
        "save_maps",
        "third_motor",
        "orientation_img_origin",
    )

    name = "grain plot"
    icon = "icons/grainplot.png"
    description = "Computes Center of mass, FWHM, Kurtosis, Skewness, mosaicity (nD), orientation (nD) maps and displays them"
    want_main_area = True
    want_control_area = False

    def __init__(self):
        super().__init__()

        self._widget = GrainPlotWidget(parent=self)
        self.mainArea.layout().addWidget(self._widget)

    def get_task_inputs(self):
        task_inputs = super().get_task_inputs()

        # Saving is handled by the widget
        task_inputs["save_maps"] = False

        return task_inputs

    def handleNewSignals(self) -> None:
        dataset = self.get_task_input_value("dataset", None)
        if dataset is not None:
            self.setDataset(dataset, pop_up=True)
        # avoid calling 'handleNewSignals' execution is already called by 'setDataset'
        # super().handleNewSignals()

    def setDataset(self, dataset: dtypes.Dataset, pop_up=False):
        if not isinstance(dataset, dtypes.Dataset):
            raise dtypes.DatasetTypeError(dataset)

        self.set_dynamic_input("dataset", dataset)
        self._widget.setMessage("Computing...")
        try:
            self.execute_ewoks_task()
        except Exception as e:
            self._widget.setMessage(f"Error while computing: {e}!")
            raise e
        if pop_up:
            self.open()

    def task_output_changed(self):
        dataset = self.get_task_output_value("dataset", None)

        if dataset is None:
            return

        if not isinstance(dataset, dtypes.Dataset):
            raise dtypes.DatasetTypeError(dataset)
        self._widget.setMessage("Computing finished!")
        self._widget.setDataset(dataset)



from __future__ import annotations

import logging
from typing import Any
from typing import Dict

import numpy
from silx.gui import qt
from silx.gui.colors import Colormap
from silx.gui.plot.StackView import StackViewMainWindow

import darfix

from ... import dtypes
from ...core.dataset import ImageDataset
from ..filterByDimension import FilterByDimensionWidget
from ..operationProcess import OperationProcess
from ..utils.message import missing_dataset_msg
from .shiftInput import ShiftInput

_logger = logging.getLogger(__file__)


class ShiftCorrectionWidget(qt.QMainWindow):
    """
    A widget to apply shift correction to a stack of images
    """

    correctSignal = qt.Signal()

    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.setWindowFlags(qt.Qt.WindowType.Widget)
        self._input_dataset: ImageDataset | None = None
        self.indices = None
        self.bg_indices = None
        self.bg_dataset = None
        self._axis = None

        self._shiftWidget = ShiftInput(parent=self)
        self._shiftWidget.correctionB.setEnabled(False)
        self._shiftWidget.correctionB.clicked.connect(self.correctSignal.emit)

        self._sv = StackViewMainWindow()
        self._sv.setColormap(
            Colormap(name=darfix.config.DEFAULT_COLORMAP_NAME, normalization="linear")
        )
        self._sv.setKeepDataAspectRatio(True)

        self._filterByDimensionWidget = FilterByDimensionWidget()

        toolbar = qt.QToolBar()
        toolbar.addWidget(self._filterByDimensionWidget)
        toolbar.addWidget(self._shiftWidget)
        self.setCentralWidget(self._sv)
        self.addToolBar(qt.Qt.ToolBarArea.RightToolBarArea, toolbar)

        self._shiftWidget.findShiftB.clicked.connect(self._launchFindShift)
        self._shiftWidget.abortShiftB.clicked.connect(self._abortFindShift)
        self._filterByDimensionWidget.axisChanged.connect(self._onAxisChanged)
        self._filterByDimensionWidget.stateDisabled.connect(
            self._onDisablingFiterByDimension
        )

        self._filterByDimensionWidget.hide()

    def getShift(self) -> numpy.ndarray:
        return numpy.array(self._shiftWidget.getShift())

    def setShift(self, shift: numpy.ndarray):
        self._shiftWidget.setShift((shift[0], shift[1]))

    def setDataset(self, dataset: dtypes.Dataset):
        """Saves the dataset and updates the stack with the dataset data."""

        self._input_dataset = dataset.dataset
        self._shiftWidget.correctionB.setEnabled(True)
        self._sv.setKeepDataAspectRatio(True)
        self._sv.setGraphTitle(self._input_dataset.title)
        self._filterByDimensionWidget.setDimensions(self._input_dataset.dims)

        self.refreshPlot()

    def _launchFindShift(self):
        dataset = self._input_dataset
        if dataset is None:
            missing_dataset_msg()
            return

        self._shiftWidget.findShiftB.setDisabled(True)
        self._shiftWidget.abortShiftB.setEnabled(True)

        self._startFindShiftThread(dataset)

    def _abortFindShift(self):
        self.thread_detection.kill()

    def _startFindShiftThread(self, dataset: ImageDataset):
        self.thread_detection = OperationProcess(
            self,
            dataset.find_shift,
            self._axis,
            indices=self.indices,
        )
        self.thread_detection.finished.connect(self._updateShift)
        self.thread_detection.start()

    def _updateShift(self, data: tuple | None):
        self._shiftWidget.findShiftB.setEnabled(True)
        self._shiftWidget.abortShiftB.setDisabled(True)
        self.thread_detection.finished.disconnect(self._updateShift)
        if data is None:
            # Aborted
            return
        self.setShift(numpy.round(data, 5))

    def getCorrectionInputs(self) -> Dict[str, Any]:
        return {
            "shift": self.getShift(),
            "selected_axis": self._axis,
        }

    def setCorrectionInputs(
        self,
        shift: tuple[float, float],
        selected_axis: int | None,
    ):
        """
        Set widget parameters from .ows save

        :param dimension_idx: dimension_idx is the index of the chosen dimension
        :param shift : shift is (shift_x, shift_y) when dimension_idx is None, else this is (shift_x1, shift_y1), ..., (shift_xn, shift_yn) with n = the dimension size
        """
        if (
            (selected_axis is not None and not isinstance(selected_axis, int))
            or not isinstance(shift, tuple)
            or not isinstance(shift[0], (int, float))
        ):
            _logger.warning(
                "Bad saved values for shift. Save file from Darfix < 4.0 ? Saved shift is discarded."
            )
            return

        self.setShift(shift)
        self._axis = selected_axis

    def _clearStack(self):
        self._sv.setStack(None)
        self._shiftWidget.correctionB.setEnabled(False)

    def _onAxisChanged(self, axis: int) -> None:
        # We only handle one fixed dimension
        self._axis = axis
        self.refreshPlot()

    def _onDisablingFiterByDimension(self) -> None:
        self._axis = None
        self.refreshPlot()

    def refreshPlot(self) -> None:
        # TODO is there another way to refresh plot ?
        if self._axis is None:
            self._sv.setStack(self._input_dataset.as_array3d(self.indices))
        else:
            self._sv.setStack(self._input_dataset.z_sum_along_axis(self._axis))



from darfix.core.roi import clampROI


def test_roi_clamping():
    """Test clamping of the ROI according to frame shape"""
    frame_origin = (0, 0)

    new_origin, new_size = clampROI(
        roi_origin=(0, 0),
        roi_size=(10, 10),
        frame_origin=frame_origin,
        frame_size=(100, 100),
    )
    assert new_origin == (0, 0)
    assert new_size == (10, 10)

    # check case ROI is fully inside the frame
    new_origin, new_size = clampROI(
        roi_origin=(0, 0),
        roi_size=(10, 10),
        frame_origin=frame_origin,
        frame_size=(3, 3),
    )
    assert new_origin == (0, 0)
    assert new_size == (3, 3)

    # check case ROI is at the upper-right corner of the frame
    new_origin, new_size = clampROI(
        roi_origin=(5, 5),
        roi_size=(10, 10),
        frame_origin=frame_origin,
        frame_size=(10, 10),
    )
    assert new_origin == (5, 5)
    assert new_size == (5, 5)

    # check case ROI is at the right of the frame
    new_origin, new_size = clampROI(
        roi_origin=(0, 5),
        roi_size=(10, 10),
        frame_origin=frame_origin,
        frame_size=(10, 10),
    )
    assert new_origin == (0, 5)
    assert new_size == (10, 5)

    # check case ROI is at the bottom of the frame
    new_origin, new_size = clampROI(
        roi_origin=(0, -5),
        roi_size=(10, 10),
        frame_origin=frame_origin,
        frame_size=(10, 10),
    )
    assert new_origin == (0, 0)
    assert new_size == (10, 5)

    # check case ROI is at the left of the frame
    new_origin, new_size = clampROI(
        roi_origin=(0, 0),
        roi_size=(10, 10),
        frame_origin=frame_origin,
        frame_size=(100, 100),
    )
    assert new_origin == (0, 0)
    assert new_size == (10, 10)



import copy

from ewokscore import Task
from ewokscore.model import BaseInputModel
from pydantic import ConfigDict

from darfix import dtypes


class Inputs(BaseInputModel):
    model_config = ConfigDict(use_attribute_docstrings=True)
    dataset: dtypes.Dataset
    """Dataset to copy."""


class DataCopy(
    Task,
    input_model=Inputs,
    output_names=["dataset"],
):
    def run(self):
        dataset = self.inputs.dataset
        if not isinstance(dataset, dtypes.Dataset):
            raise dtypes.DatasetTypeError(dataset)
        self.outputs.dataset = dtypes.Dataset(
            dataset=copy.deepcopy(dataset.dataset),
            indices=copy.deepcopy(dataset.indices),
            bg_indices=copy.deepcopy(dataset.bg_indices),
            bg_dataset=copy.deepcopy(dataset.bg_dataset),
        )



from __future__ import annotations

import logging
from typing import Tuple

import h5py

_logger = logging.getLogger(__file__)


def find_scan_names(file: str) -> Tuple[str, ...]:
    with h5py.File(file) as h5file:
        return tuple(
            entry.name.lstrip("/")
            for entry in h5file.values()
            if isinstance(entry, h5py.Group)
        )


def find_detector_name(scan: h5py.Group) -> str | None:
    instrument_group = scan.get("instrument", None)
    if not isinstance(instrument_group, h5py.Group):
        _logger.warning(
            f"Could not find group 'instrument' in {scan.file}::{scan.name}"
        )
        return None

    for instrument_name, instrument in instrument_group.items():
        if not isinstance(instrument, h5py.Group):
            continue

        if instrument.attrs.get("NX_class") != "NXdetector":
            continue

        instrument_type = instrument.get("type")
        if not isinstance(instrument_type, h5py.Dataset):
            continue
        if instrument_type[()].decode() == "lima" and "image" in instrument:
            # Found the LIMA detector
            return instrument_name

    _logger.warning(
        f"Could not find valid detector in {scan.file}::{instrument_group.name} among {tuple(instrument_group.keys())}"
    )
    return None



from __future__ import annotations

from concurrent.futures import ThreadPoolExecutor
from concurrent.futures import as_completed
from typing import Any
from typing import Callable
from typing import TypedDict

import numpy
from silx.utils.enum import Enum as _Enum
from tqdm import tqdm

from darfix.core.utils import OperationAborted

from ..dtypes import Dataset
from ..processing import imageOperations
from .state_of_operation import Operation


class BackgroundType(_Enum):
    DATA = "Data"
    UNUSED_DATA = "Unused data (after partition)"
    DARK_DATA = "Dark data"


class NoiseRemovalOperation(TypedDict):
    type: Operation
    parameters: dict[str, Any]
    aborted: bool


def operation_to_str(op: NoiseRemovalOperation) -> str:
    result = None
    if op["type"] is Operation.BS:
        result = f"Background subtraction {op['parameters']}"

    elif op["type"] is Operation.HP:
        result = f"Hot pixel removal: {op['parameters']}"

    elif op["type"] is Operation.THRESHOLD:
        result = f"Threshold removal: {op['parameters']}"

    elif op["type"] is Operation.MASK:
        result = "Mask removal"

    if result is not None:
        return result + (" ABORTED" if "aborted" in op else "")

    return super().__str__(op)


def apply_noise_removal_operation(
    image: numpy.ndarray, operation: NoiseRemovalOperation
) -> None:
    if operation["type"] is Operation.BS:
        apply_background_substraction(image, **operation["parameters"])

    if operation["type"] is Operation.HP:
        apply_hot_pixel_removal(image, **operation["parameters"])

    if operation["type"] is Operation.THRESHOLD:
        apply_threshold_removal(image, **operation["parameters"])

    if operation["type"] is Operation.MASK:
        apply_mask_removal(image, **operation["parameters"])


def create_background_substraction_operation(
    dataset: Dataset,
    method: str | None = None,
    background_type: BackgroundType | None = None,
) -> NoiseRemovalOperation:
    """Compute a background image and return a NoiseRemovalOperation with a background parameter"""

    if method is None:
        method = imageOperations.Method.MEDIAN

    if background_type is not None:
        background_type = BackgroundType.from_value(background_type)

    if background_type == BackgroundType.DARK_DATA:
        bg = dataset.bg_dataset.as_array3d()
    elif background_type == BackgroundType.UNUSED_DATA:
        bg = dataset.dataset.as_array3d(dataset.bg_indices)
    elif background_type == BackgroundType.DATA:
        bg = dataset.dataset.as_array3d()
    else:
        raise NotImplementedError(
            f"Background type {background_type!r} not implemented yet."
        )

    return NoiseRemovalOperation(
        type=Operation.BS,
        parameters=dict(background=imageOperations.compute_background(bg, method)),
    )


def apply_background_substraction(
    image: numpy.ndarray, background: numpy.ndarray
) -> None:
    imageOperations.background_subtraction(image, background)


def apply_hot_pixel_removal(
    image: numpy.ndarray, kernel_size: int | None = None
) -> None:
    if kernel_size is None:
        kernel_size = 3

    imageOperations.hot_pixel_removal(image, ksize=kernel_size)


def apply_threshold_removal(
    image: numpy.ndarray, bottom: int | None = None, top: int | None = None
) -> None:
    imageOperations.threshold_removal(image, bottom=bottom, top=top)


def apply_mask_removal(image: numpy.ndarray, mask: numpy.ndarray | None) -> None:
    if mask is None:
        return
    imageOperations.mask_removal(image, mask)


def apply_noise_removal_operations(
    image_stack: numpy.ndarray,
    operations: list[NoiseRemovalOperation],
    is_cancel: Callable[[], bool],
    set_progress: Callable[[int], None],
    max_workers: int | None = None,
) -> None:
    if image_stack.ndim != 3:
        raise ValueError(
            f"image_stack must have ndim == 3 but ndims is {image_stack.ndim}"
        )

    def process_single(img: numpy.ndarray):
        if is_cancel():
            return
        for op in operations:
            try:
                apply_noise_removal_operation(img, operation=op)
            except Exception as e:
                # Return e to re-raise Exception in the parent thread
                return e

    with ThreadPoolExecutor(max_workers=max_workers) as executor:
        futures = [executor.submit(process_single, img) for img in image_stack]
        len_futures = len(futures)
        for idx_future, future in tqdm(
            enumerate(as_completed(futures)),
            desc="Noise removal operation",
            total=len_futures,
        ):
            exception = future.result()
            if isinstance(exception, Exception):
                raise exception
            if is_cancel():
                raise OperationAborted()
            set_progress(idx_future * 100 // len_futures)



from typing import Optional

from ewokscore.missing_data import MISSING_DATA
from ewoksorange.bindings.owwidgets import OWEwoksWidgetOneThread
from ewoksorange.gui.orange_imports import Output
from ewoksorange.gui.orange_imports import Setting

from darfix import dtypes
from darfix.gui.blindSourceSeparationWidget import BSSWidget
from darfix.tasks.blindsourceseparation import BlindSourceSeparation
from darfix.tasks.blindsourceseparation import Method


class BlindSourceSeparationWidgetOW(
    OWEwoksWidgetOneThread, ewokstaskclass=BlindSourceSeparation
):
    """
    Widget to apply blind source separation (BSS) to find grains along the dataset.
    Several techniques can be used like NMF, NNICA, and NMF+NNICA.
    """

    name = "blind source separation"
    icon = "icons/bss.png"
    want_main_area = False

    _ewoks_inputs_to_hide_from_orange = ("method", "n_comp", "processing_order", "save")
    _ewoks_outputs_to_hide_from_orange = ("W", "comp")

    # Settings
    method = Setting(BSSWidget.DEFAULT_METHOD, schema_only=True)
    n_comp = Setting(BSSWidget.DEFAULT_N_COMP, schema_only=True)

    # Outputs
    class Outputs:
        dataset = Output("dataset", dtypes.Dataset)
        comp = Output("comp", object)
        W = Output("W", object)

    def __init__(self):
        super().__init__()
        # backward compatibility
        # if a previous workflow or settings cache contains invalid values, clean them
        self._widget = BSSWidget(parent=self)
        self._widget.sigComputed.connect(self._updateSettings)
        self.controlArea.layout().addWidget(self._widget)

        # set up the gui
        self._widget.setMethod(self.method)
        self._widget.setNComp(self.n_comp)
        # safer: in the case of having invalid settings store updating then will make sure
        # they get valid values
        self._updateSettings(
            method=self._widget.getMethod(),
            n_comp=self._widget.getNComp(),
        )

        # connect signal / slot
        self._widget.computeButton.released.connect(self.execute_ewoks_task)
        self._widget.sigNbComponentsChanged.connect(self._nbComponentChanged)
        self._widget.sigMethodChanged.connect(self._methodChanged)

    def _updateSettings(self, method, n_comp):
        self._nbComponentChanged(n_comp)
        self._methodChanged(method=method)

    def setDataset(self, dataset: Optional[dtypes.Dataset], pop_up=False):
        if dataset in (None, MISSING_DATA):
            return

        if not isinstance(dataset, dtypes.Dataset):
            raise TypeError(
                f"dataset is expected to be an instance of Dataset. Get {type(dataset)}"
            )
        self._widget.setDataset(dataset)

        if pop_up:
            self.open()

    def _updateDataset(self, widget, dataset):
        self._widget._updateDataset(widget, dataset)

    def _nbComponentChanged(self, nb_components: int):
        self.n_comp = nb_components

    def _methodChanged(self, method: str):
        self.method = Method.from_value(method)

    def handleNewSignals(self) -> None:
        """Invoked by the workflow signal propagation manager after all
        signals handlers have been called.

        note: this widget can receive one signal: 'dataset'. The 'dataset' signal is handled by the ewoks task.
              This function will be only triggered when the 'dataset' signal is send
        """
        # update GUI from received dataset
        dataset = self.get_task_input_value("dataset")
        self.setDataset(dataset=dataset, pop_up=True)

        super().handleNewSignals()

    def get_task_inputs(self):
        task_inputs = super().get_task_inputs()

        task_inputs.update(
            {
                "method": self.method,
                "n_comp": self.n_comp,
            }
        )
        return task_inputs

    def task_output_changed(self):
        # once the output is computed we can update the GUI (calling widget._displayComponents)
        comp = self.get_task_output_value("comp", MISSING_DATA)
        W = self.get_task_output_value("W", MISSING_DATA)
        dataset = self.get_task_output_value("dataset", MISSING_DATA)
        if (
            comp is not MISSING_DATA
            and W is not MISSING_DATA
            and dataset is not MISSING_DATA
        ):
            if not isinstance(dataset, dtypes.Dataset):
                raise dtypes.DatasetTypeError(dataset)
            self._widget._displayComponents(
                dataset=dataset.dataset,
                indices=dataset.indices,
                bg_indices=dataset.bg_indices,
                comp=comp,
                W=W,
            )
        super().task_output_changed()



from silx.utils.enum import Enum as _Enum


class ConfigurationLevel(_Enum):
    REQUIRED = "required"
    OPTIONAL = "optional"
    ADVANCED = "advanced"

    def _get_num_value(self) -> int:
        if self is self.REQUIRED:
            return 0
        elif self is self.OPTIONAL:
            return 1
        elif self is self.ADVANCED:
            return 2

    def __le__(self, other):
        if not isinstance(other, ConfigurationLevel):
            raise TypeError(
                f"other is expected to be an instance of ConfigurationLevel. {type(other)} provided"
            )
        return self._get_num_value() <= other._get_num_value()

    def __lt__(self, other):
        if not isinstance(other, ConfigurationLevel):
            raise TypeError(
                f"other is expected to be an instance of ConfigurationLevel. {type(other)} provided"
            )
        return self._get_num_value() < other._get_num_value()



from silx.gui import qt

if qt.BINDING == "PyQt5":
    from PyQt5.QtTest import QSignalSpy
elif qt.BINDING == "PySide6":
    from PySide6.QtTest import QSignalSpy
elif qt.BINDING == "PyQt6":
    from PyQt6.QtTest import QSignalSpy
else:
    QSignalSpy = None



from __future__ import annotations

import logging

import numpy
from silx.gui import qt

from darfix.core.dimension import Dimension
from darfix.tasks.dimensiondefinition import DimensionDefinition

_logger = logging.getLogger(__file__)


class DimensionTable(qt.QTableWidget):
    """
    Widget used to define the number of dimension and with which values they are
    mapped
    """

    sigUpdateDims = qt.Signal()
    """
    If dim added, removed or size modified
    """

    _V_HEADERS = ["Axis", "Name", "Size", "Start", "Stop", ""]

    def __init__(self, parent: qt.QWidget | None = None):
        super().__init__(parent)

        self.setColumnCount(len(self._V_HEADERS))
        self.setHorizontalHeaderLabels(self._V_HEADERS)
        header = self.horizontalHeader()
        header.setMinimumSectionSize(50)
        for iColumn in range(len(self._V_HEADERS)):
            header.setSectionResizeMode(iColumn, qt.QHeaderView.Stretch)
        self.verticalHeader().hide()
        self._dims: dict[int, DimensionItem] = {}

    def clear(self):
        self._dims = {}
        self.clearContents()
        self.setRowCount(0)

    @property
    def ndim(self) -> int:
        return len(self._dims)

    @property
    def dims(self) -> dict[int, DimensionItem]:
        return self._dims

    def getShape(self) -> tuple[int, ...]:
        return tuple(dimItem.size for dimItem in self.dims.values())

    def _addDim(self, axis: int, dim: Dimension, decimals: int):
        """

        :param axis: which axis is defining this dimension
        :param `Dimension` dim: definition of the dimension to add
        """
        axis = self._getNextFreeAxis()
        row = self.rowCount()
        self.setRowCount(row + 1)
        widget = DimensionItem(parent=self, table=self, row=row)
        widget.setStartStopDecimals(decimals)
        widget.sigRemoved.connect(self.onRemoveDim)
        widget.sigDimChanged.connect(self.sigUpdateDims)
        widget.fromDimension(axis, dim)
        self._dims[row] = widget
        return widget

    def setDims(self, dims: dict[int, Dimension], decimals: int):
        """

        :param dict dims: axis as key and `Dimension` as value.
        """
        self.clear()
        if not isinstance(dims, dict):
            raise TypeError(f"dims should be a dict. Got {type(dims)}")

        for axis, dim in dims.items():
            assert type(axis) is int
            assert isinstance(dim, Dimension)
            self._addDim(axis, dim, decimals)

        self.sigUpdateDims.emit()

    def onRemoveDim(self, iRow: int):
        """
        Remove dimension.

        :param Union[int,`DimensionItem`]: row or item to remove
        """
        # remove the widget
        self.removeRow(iRow)
        self._dims[iRow].sigRemoved.disconnect(self.onRemoveDim)
        self._dims[iRow].sigDimChanged.disconnect(self.sigUpdateDims)
        self._dims[iRow].setAttribute(qt.Qt.WA_DeleteOnClose)
        self._dims[iRow].close()
        del self._dims[iRow]
        # reorder existing widgets
        ini_rows = sorted(list(self._dims.keys()))
        for row in ini_rows:
            if row <= iRow:
                continue
            widget = self._dims[row]
            new_row = row - 1
            assert new_row >= 0
            widget.embedInTable(table=self, row=new_row)
            widget.axis = self._getNextFreeAxis()
            self._dims[new_row] = widget
            del self._dims[row]
        self.sigUpdateDims.emit()

    def getDim(self, iRow: int):
        """return the instance of DimensionItem for the iRow"""
        return self._dims.get(iRow, None)

    def _getNextFreeAxis(self):
        """
        :return int: next unused axis
        """
        res = 0
        usedAxis = []
        [usedAxis.append(_dim.axis) for _dim in self._dims.values()]
        while res in usedAxis:
            res = res + 1
        return res


class DimensionWidget(qt.QWidget):
    """
    Widget to define dimensions and try to fit those with dataset
    """

    sigUpdateDimensions = qt.Signal()
    """Emitted when user inputs in the dimension table changed."""
    sigFindDimensions = qt.Signal()
    """Emit when user want to find dimensions"""

    def __init__(self, parent=None):
        super().__init__(parent)

        # ============ SCAN SETTINGS ============
        zigzagModeLabel = qt.QLabel("Zigzag Mode: ")
        self._isZigzagCB = qt.QCheckBox()
        self._isZigzagCB.setToolTip(
            "Checked if acquisition of scan with bliss is in zig zag mode.\n Fill by 'find dimensions' button."
        )

        # ============ FIND DIMS ============

        toleranceLabel = qt.QLabel("Tolerance = 1e-")
        self._toleranceSB = qt.QSpinBox()
        self._toleranceSB.valueChanged.connect(self.onToleranceChanged)

        self._toleranceSB.setToolTip(
            "Tolerance for which the values in the dimensions will be considered unique"
        )

        self._findDimsButton = qt.QPushButton("Find dimensions")
        self._findDimsButton.setToolTip(
            "Automatically find all dimensions that "
            "change through the dataset. \n"
            "It is considered that a dimension changes "
            "if the number of unique values is greater "
            "than one.\nThe metadata type is needed to "
            "choose which values to compare, and the "
            "threshold to know when two values are "
            "considered to be different.\nThe threshold"
            " is only used in values that are numbers."
        )

        # ============ DIMS TABLE ============

        self._dimensionTable = DimensionTable(self)
        self._dimensionTable.sigUpdateDims.connect(self.sigUpdateDimensions)

        # ============ TIPS LABEL ============

        self._tipsLabel = qt.QLabel(self)

        # ============ LAYOUT EVERYTHING ============

        detection_layout = qt.QHBoxLayout()
        detection_layout.setAlignment(
            qt.Qt.AlignmentFlag.AlignLeft | qt.Qt.AlignmentFlag.AlignTop
        )
        detection_layout.addWidget(toleranceLabel)
        detection_layout.addWidget(self._toleranceSB)
        detection_layout.addWidget(self._findDimsButton)

        dimensions_definition_group = qt.QGroupBox("Dimensions definition")
        dimensions_definition_group.setLayout(qt.QVBoxLayout())
        dimensions_definition_group.layout().addWidget(self._dimensionTable)
        dimensions_definition_group.layout().addLayout(detection_layout)

        scan_settings_group = qt.QGroupBox("Scan settings")
        scan_settings_group.setLayout(qt.QHBoxLayout())
        scan_settings_group.layout().setAlignment(
            qt.Qt.AlignmentFlag.AlignLeft | qt.Qt.AlignmentFlag.AlignTop
        )
        scan_settings_group.layout().addWidget(zigzagModeLabel)
        scan_settings_group.layout().addWidget(self._isZigzagCB)

        main_layout = qt.QVBoxLayout()
        main_layout.addWidget(scan_settings_group)
        main_layout.addWidget(dimensions_definition_group)
        main_layout.addWidget(self._tipsLabel)

        self.setLayout(main_layout)

        # connect Signal/SLOT
        self._findDimsButton.pressed.connect(self._findDimensions)
        self.setTolerance(DimensionDefinition.DEFAULT_TOLERANCE)

    def getTolerance(self) -> float:
        # From N to 1e-N
        return 10 ** (-self._toleranceSB.value())

    def setTolerance(self, tolerance: float):
        # From 1e-N to N
        self._toleranceSB.setValue(-int(numpy.log10(tolerance)))

    def onToleranceChanged(self, val: int):
        for item in self._dimensionTable.dims.values():
            item.setStartStopDecimals(val)

    def _findDimensions(self):
        """
        Try to find dimensions
        """
        self.sigFindDimensions.emit()

    def setTipsLabelText(self, txt: str):
        self._tipsLabel.setText(txt)

    def setDims(self, dims: dict[int, Dimension]):
        self._dimensionTable.setDims(dims, self._toleranceSB.value())

    @property
    def ndim(self) -> int:
        return self._dimensionTable.ndim

    @property
    def dims(self) -> dict[int, DimensionItem]:
        return self._dimensionTable.dims

    @property
    def isZigzagMode(self):
        return self._isZigzagCB.isChecked()

    def setZigzagMode(self, isZigzag: bool):
        return self._isZigzagCB.setChecked(isZigzag)

    def setEnableInputs(self, enabled: bool):
        """Enable/disable controls"""
        self._isZigzagCB.setEnabled(enabled)
        self._findDimsButton.setEnabled(enabled)
        self._toleranceSB.setEnabled(enabled)
        for dimItem in self._dimensionTable.dims.values():
            dimItem.setEnabledInputs(enabled)


class DimensionItem(qt.QWidget):
    """Widget use to define a dimension"""

    sigRemoved = qt.Signal(int)
    """Signal emitted when the Item should be removed"""

    sigDimChanged = qt.Signal()
    """Signal emitted when dim changed"""

    def __init__(self, parent, table, row):
        """

        :param QTableWidget table: if has to be embed in a table the
                                           parent table
        :param int row: row position in the QTableWidget. Also used as ID
        """
        qt.QWidget.__init__(self, parent)

        # axis
        self._axisLabel = qt.QLabel(parent=self)
        self._axisLabel.setAlignment(
            qt.QtCore.Qt.AlignCenter | qt.QtCore.Qt.AlignHCenter
        )
        # name
        self._nameLabel = qt.QLabel(parent=self)
        self._nameLabel.setAlignment(
            qt.QtCore.Qt.AlignCenter | qt.QtCore.Qt.AlignHCenter
        )
        # size
        self._sizeWidget = qt.QSpinBox(self)
        self._sizeWidget.setMinimum(0)
        self._sizeWidget.setMaximum(9999)
        self._sizeWidget.valueChanged.connect(self.sigDimChanged)
        # start
        self._startWidget = qt.QDoubleSpinBox(self)
        self._startWidget.setMinimum(float("-inf"))
        self._startWidget.setDecimals(5)
        self._startWidget.valueChanged.connect(self.sigDimChanged)
        # end
        self._stopWidget = qt.QDoubleSpinBox(self)
        self._stopWidget.setMinimum(float("-inf"))
        self._stopWidget.setDecimals(5)
        self._stopWidget.valueChanged.connect(self.sigDimChanged)
        # rm button
        style = qt.QApplication.style()
        icon = style.standardIcon(qt.QStyle.SP_BrowserStop)
        self._rmButton = qt.QPushButton(icon=icon, parent=self)

        # connect Signal/slot
        self._rmButton.pressed.connect(self.remove)

        self.embedInTable(table=table, row=row)
        self.__row = row

    def setEnabledInputs(self, enabled: bool):
        self._sizeWidget.setEnabled(enabled)
        self._startWidget.setEnabled(enabled)
        self._stopWidget.setEnabled(enabled)
        self._rmButton.setEnabled(enabled)

    def remove(self):
        self.sigRemoved.emit(self._row)

    def setStartStopDecimals(self, decimals: int):
        self._startWidget.setDecimals(decimals)
        self._stopWidget.setDecimals(decimals)

    @property
    def _row(self):
        return self.__row

    @property
    def axis(self):
        return int(self._axisLabel.text())

    @axis.setter
    def axis(self, axis: int):
        self._axisLabel.setText(str(axis))

    @property
    def size(self):
        return self._sizeWidget.value()

    @property
    def start(self):
        return self._startWidget.value()

    @property
    def stop(self):
        return self._stopWidget.value()

    @property
    def name(self):
        return self._nameLabel.text()

    def toDimension(self) -> Dimension:
        return Dimension(self.name, self.size, self.start, self.stop)

    def fromDimension(self, axis: int, dim: Dimension):
        self.axis = axis
        self._nameLabel.setText(dim.name)
        self._sizeWidget.setValue(dim.size)
        self._startWidget.setValue(dim.start)
        self._stopWidget.setValue(dim.stop)

    def embedInTable(self, table, row):
        self.__row = row
        for column, widget in enumerate(
            (
                self._axisLabel,
                self._nameLabel,
                self._sizeWidget,
                self._startWidget,
                self._stopWidget,
                self._rmButton,
            )
        ):
            table.setCellWidget(row, column, widget)



class Config:
    """
    Class containing shared global configuration for the darfix project.

    .. versionadded:: 0.3
    """

    DEFAULT_COLORMAP_NAME = "cividis"
    DEFAULT_COLORMAP_NORM = "linear"
    """Default LUT for the plot widgets.

    The available list of names are available in the module
    :module:`silx.gui.colors`.

    .. versionadded:: 0.3
    """

    FWHM_VAL = 2.35482

    """Magic value that returns FWHM when multiplied by the standard deviation.

    .. versionadded:: 0.5
    """



import warnings
from pathlib import Path

import h5py
import numpy
from scipy.optimize import OptimizeWarning

from darfix.core.rocking_curves import Maps_2D
from darfix.dtypes import Dataset
from darfix.tasks.rocking_curves import RockingCurves


def test_rocking_curves(input_dataset, tmp_path):
    output_filename = tmp_path / "rocking_curves.h5"
    input_dataset.find_dimensions()
    input_dataset.reshape_data()
    task = RockingCurves(
        inputs={
            "dataset": Dataset(input_dataset),
            "method": "lm",
            "output_filename": output_filename,
            "int_thresh": 15,
        }
    )
    with warnings.catch_warnings():
        warnings.simplefilter("always", RuntimeWarning)
        warnings.simplefilter("always", OptimizeWarning)
        task.run()

    assert isinstance(task.get_output_value("maps"), numpy.ndarray)
    assert isinstance(task.get_output_value("dataset"), Dataset)
    assert output_filename.is_file()
    with h5py.File(output_filename, "r") as output_file:
        output_entry = output_file["entry"]
        for map_name in Maps_2D.values():
            assert map_name in output_entry
        assert "Residuals" in output_entry


def test_rocking_curves_no_save(input_dataset):
    input_dataset.find_dimensions()
    input_dataset.reshape_data()
    task = RockingCurves(
        inputs={
            "dataset": Dataset(input_dataset),
            "method": "lm",
            "output_filename": None,
            "int_thresh": 15,
        }
    )
    with warnings.catch_warnings():
        warnings.simplefilter("always", RuntimeWarning)
        warnings.simplefilter("always", OptimizeWarning)
        task.run()

    assert isinstance(task.get_output_value("maps"), numpy.ndarray)
    assert isinstance(task.get_output_value("dataset"), Dataset)
    default_filename = Path(input_dataset._dir) / "rocking_curves.h5"
    assert not default_filename.is_file()



from __future__ import annotations

from typing import Callable

from ewoksorange.gui.parameterform import block_signals
from silx.gui import qt

from darfix.core.dimension import AcquisitionDims

from ..dtypes import AxisAndValueIndices


class _DimensionWidget(qt.QWidget):
    def __init__(
        self,
        parent,
        value_callback: Callable,
        show_values: bool | None = None,
    ) -> None:
        super().__init__(parent)
        layout = qt.QGridLayout()
        self.dimensionCB = qt.QComboBox()
        self.valueCB = None
        if show_values:
            dimensionLabel = qt.QLabel("Dimension: ")
            layout.addWidget(dimensionLabel, 0, 0)
            layout.addWidget(self.dimensionCB, 0, 1)
            valueLabel = qt.QLabel("Value: ")
            self.valueCB = qt.QComboBox()
            layout.addWidget(valueLabel, 1, 0)
            layout.addWidget(self.valueCB, 1, 1)
            self.valueCB.currentIndexChanged.connect(value_callback)
        else:
            layout.addWidget(self.dimensionCB, 0, 0)
        self.setLayout(layout)
        self.setEnabled(False)

    @property
    def dimensionComboBox(self):
        return self.dimensionCB

    @property
    def valueComboBox(self):
        return self.valueCB


class ChooseDimensionDock(qt.QDockWidget):
    def __init__(self, parent=None, vertical=True, values=True, _filter=True):
        """
        Dock widget containing the ChooseDimensionWidget.
        """
        qt.QDockWidget.__init__(self, parent)
        self.setWidget(ChooseDimensionWidget(self, vertical, values, _filter))

    def widget(self) -> ChooseDimensionWidget:
        child = super().widget()
        assert isinstance(child, ChooseDimensionWidget), type(child)
        return child


class ChooseDimensionWidget(qt.QWidget):
    """
    Widget to choose a dimension from a dict and choose the value to filter
    the data. It can be included in other widget like StackView to filter the
    stack.
    """

    valueChanged = qt.Signal()
    filterChanged = qt.Signal(list, list)
    """ Signal sending the new filter dimension and value. Only emitted if filter is true."""
    stateDisabled = qt.Signal()

    def __init__(self, parent=None, vertical=True, values=True, _filter=True):
        qt.QWidget.__init__(self, parent)

        if vertical:
            self.setLayout(qt.QVBoxLayout())
        else:
            self.setLayout(qt.QHBoxLayout())
        self.value: list[int] = []
        self.show_values: bool = values
        self.dimensionWidgets: list[_DimensionWidget] = []
        self.dimension: list[int] = []
        self.dimensions: AcquisitionDims | None = None
        self.filter = _filter
        if _filter:
            self._checkbox = qt.QCheckBox("Filter by dimension", self)
            self._checkbox.toggled.connect(self._updateState)

    def isFilteredByDim(self) -> bool:
        return self._checkbox.isChecked()

    def setFilterByDim(self, filter: bool):
        return self._checkbox.setChecked(filter)

    def setDimensions(
        self,
        dimensions: AcquisitionDims,
        selectedDimAxis: int = 0,
        seletedDimIndex: int = 0,
    ):
        """
        Function that fills the corresponding comboboxes with the dimension's
        name and possible values.

        :param array_like dimensions: List of `darfix.core.dataset.Dimension`
                                      elements.
        """
        with block_signals(
            self
        ):  # Do not fire valueChanged or filterChanged when setting up the widget
            self.dimensionWidgets = []
            for i in reversed(range(self.layout().count())):
                self.layout().itemAt(i).widget().setParent(None)
            self.dimensions = dimensions
            self.dimension = []
            self.value = [0 for i in range(self.dimensions.ndim - 1)]
            for i in range(dimensions.ndim - 1):
                self._addDimensionWidget()
                self.dimension.append(i)
                self.dimensionWidgets[-1].dimensionComboBox.setCurrentIndex(i)

            self._updateDimension(selectedDimAxis, seletedDimIndex)
            if self.filter:
                self._updateState(self._checkbox.isChecked())
                self.layout().addWidget(self._checkbox)

    def _addDimensionWidget(self):
        """
        Add new widget to choose between different dimensions and values.
        """

        widget = _DimensionWidget(self, self._updateValue, self.show_values)
        dimensionCB = widget.dimensionComboBox

        assert self.dimensions is not None
        for axis, dimension in self.dimensions.items():
            dimensionCB.insertItem(axis, dimension.name)
        dimensionCB.currentIndexChanged.connect(self._updateDimension)

        self.layout().addWidget(widget)
        self.dimensionWidgets.append(widget)

    def _updateDimension(self, current_axis=-1, seletedDimIndex: int = 0):
        """
        Updates the selected dimension and set's the corresponding possible values.

        :param int axis: selected dimension's axis, only used to check valid call
            of the method.
        """
        if current_axis == -1 or current_axis is None:
            return

        assert self.dimensions is not None
        self.dimension = []
        # Init values to 0
        self.value = [0 for i in range(self.dimensions.ndim - 1)]
        # Reset all dimensions
        for dimWidget in self.dimensionWidgets:
            # Prevent signals
            dimCB = dimWidget.dimensionComboBox
            with block_signals(dimCB):
                current_axis = dimCB.currentIndex()
                # Enable / disable items in combobox
                for axis in self.dimensions.keys():
                    if axis in self.dimension:
                        dimCB.model().item(axis).setEnabled(False)
                        # If axis is already in the dimensions list, set it to
                        # next available axis.
                        if current_axis == axis:
                            current_axis = (current_axis + 1) % self.dimensions.ndim
                            dimCB.setCurrentIndex(current_axis)
                    else:
                        dimCB.model().item(axis).setEnabled(True)
            if dimCB.currentText() != "None":
                self.dimension.append(current_axis)
            valueCB = dimWidget.valueComboBox
            if valueCB is not None:
                # Set values from new dimension
                valueCB.clear()
                valueCB.addItems(
                    map(str, self.dimensions.get(current_axis).compute_unique_values())
                )
                valueCB.setCurrentIndex(seletedDimIndex)

        self.valueChanged.emit()
        if self.filter:
            self.filterChanged.emit(self.dimension, self.value)

    def _updateValue(self, index=None):
        """
        Updates the selected value.

        :param int index: selected value's index.
        """
        if self.show_values and (index is not None or index != -1):
            self.value = []
            for dimWidget in self.dimensionWidgets:
                axis = dimWidget.valueComboBox.currentIndex()
                self.value.append(axis)

            self.valueChanged.emit()
            if self.filter:
                self.filterChanged.emit(self.dimension, self.value)

    def _updateState(self, checked: bool):
        """
        Updates the state of the widget.

        :param checked: If True, the widget emits a signal
                    with the selected dimension and value. Else,
                    a disabled signal is emitted.

        """
        for dimWidget in self.dimensionWidgets:
            dimWidget.setEnabled(checked)

        if checked:
            self.filterChanged.emit(self.dimension, self.value)
        else:
            self.stateDisabled.emit()

    def setDimension(self, dimension: AxisAndValueIndices):
        axis_indices, value_indices = dimension
        for dimWidget, axis_idx, value_idx in zip(
            self.dimensionWidgets, axis_indices, value_indices
        ):

            # Do not fire signals to avoid multiple computation triggers
            # See https://gitlab.esrf.fr/XRD/darfix/-/merge_requests/446
            with block_signals(dimWidget):
                dimWidget.dimensionComboBox.setCurrentIndex(axis_idx)
                if dimWidget.valueComboBox is not None:
                    dimWidget.valueComboBox.setCurrentIndex(value_idx)

        self._checkbox.setChecked(True)
